#!/usr/bin/env python
#
# program.py - The GLSLShader class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLSLShader` class, which encapsulates
a GLSL shader program comprising a vertex shader, a fragment shader, and
optionally a geometry shader (for OpenGL >= 3.3).
"""


import logging
import contextlib

import jinja2                as j2
import numpy                 as np
import OpenGL.GL             as gl
import OpenGL.raw.GL._types  as gltypes

import fsleyes.gl.extensions as glexts
import fsleyes.gl.resources  as glresources
import fsleyes.gl            as fslgl
import fsl.utils.memoize     as memoize
from . import                   parse


log = logging.getLogger(__name__)


GLSL_ATTRIBUTE_TYPES = {
    'bool'  : (gl.GL_BOOL,  1),
    'int'   : (gl.GL_INT,   1),
    'float' : (gl.GL_FLOAT, 1),
    'vec2'  : (gl.GL_FLOAT, 2),
    'vec3'  : (gl.GL_FLOAT, 3),
    'vec4'  : (gl.GL_FLOAT, 4)
}
"""This dictionary contains mappings between GLSL data types, and their
corresponding GL types and sizes.
"""


class GLSLShader:
    """The ``GLSLShader`` class encapsulates information and logic about a GLSL
    shader program, comprising a vertex shader, a fragment shader, and
    optionally a geometry shader (if OpenGL >= 3.3 is available). It provides
    methods to set shader attribute and uniform values, to configure
    attributes, and to load/unload the program. Furthermore, the
    ``GLSLShader`` makes sure that all uniform and attribute variables are
    converted to the appropriate type. The following methods are available on
    a ``GLSLShader``:


    .. autosummary::
       :nosignatures:

       load
       unload
       loaded
       destroy
       set
       setAtt
       setIndices
       ready
       draw


    Typical usage of a ``GLSLShader`` will look something like the
    following::

        vertSrc = 'vertex shader source'
        fragSrc = 'fragment shader source'

        program = GLSLShader(vertSrc, fragSrc)

        # Load the program
        with shader.loaded():

            # Set some uniform values
            program.set('lighting', True)
            program.set('lightPos', [0, 0, -1])

            # Create and set vertex attributes
            vertices, normals = createVertices()

            program.setAtt('vertex', vertices)
            program.setAtt('normal', normals)

            # Draw the scene
            program.draw(gl.GL_TRIANGLES, len(vertices))

        # Delete the program when
        # we no longer need it
        program.destroy()
    """


    def __init__(self,
                 vertSrc,
                 fragSrc,
                 geomSrc=None,
                 constants=None,
                 resourceName=None,
                 shared=None):
        """Create a ``GLSLShader``.

        The source is passed through ``jinja2``, replacing any expressions on
        the basis of ``constants``.

        :arg vertSrc:      String containing vertex shader source code.

        :arg fragSrc:      String containing fragment shader source code.

        :arg geomSrc:      String containing geometry shader source code.

        :arg constants:    Key-value pairs to be used when passing the source
                           through ``jinja2``.

        :arg resourceName: If provided, buffers for any ``shared`` attributes
                           will be managed using the :mod:`.resources` module.

        :arg shared:       Names of input/varying vertex attributes which are
                           shared between multiple shaders - if a
                           ``resourceName`` is provided, these attributes will
                           be mamaged by the :mod:`.resources` module.
        """

        if shared is None:
            shared = []

        if constants is None:
            constants = {}

        srcs = []
        for src in (vertSrc, fragSrc, geomSrc):
            if src is not None:
                srcs.append(j2.Template(src).render(**constants))
        types      = {}
        sizes      = {}
        attributes = []
        uniforms   = set()

        # Extract vertex and constant inputs
        # required by the shader program -
        # anything from the vertex shader
        # declared as 'varying' or 'in', and
        # anything from any of the shaders
        # declared as 'uniform'.
        for i, src in enumerate(srcs):
            if src is None:
                continue
            decs = parse.parseGLSL(src)

            # get attributes/vertex inputs from
            # vertex shader. For the other shaders,
            # we only care about uniforms.
            if i == 0:
                atts  = decs['attribute']
                unifs = decs['uniform']
            else:
                atts  = []
                unifs = decs['uniform']

            attributes.extend(atts)
            uniforms = uniforms.union(unifs)
            for dname, dtype, dsize in (atts + unifs):
                types[dname] = dtype
                sizes[dname] = dsize

        self.program     = self.__compile(*srcs)
        self.attDivisors = {}
        self.types       = types
        self.sizes       = sizes
        self.attributes  = [a[0] for a in attributes]
        self.uniforms    = [u[0] for u in uniforms]
        self.positions   = self.__getPositions(self.program,
                                               self.attributes,
                                               self.uniforms)

        # Flags indicating whether each uniform/attribute
        # has been given a value via set/setAtt
        self.hasValue = {a : False for a in self.attributes + self.uniforms}

        # Buffers for vertex attributes
        self.buffers = {}
        for att in self.attributes:
            if resourceName is not None and att in shared:
                arname = f'{resourceName}_{att}'
                self.buffers[att] = glresources.get(arname, gl.glGenBuffers, 1)
            else:
                self.buffers[att] = gl.glGenBuffers(1)

        # Buffers for storing vertices and
        # (optionally) vertex indices.
        # A vertex array is required in
        # GL >= 3.0, but not supported in
        # older GL versions.
        #
        # A vertex index buffer is created
        # on the first call to setIndices.
        if float(fslgl.GL_COMPATIBILITY) >= 3:
            self.vao = gl.glGenVertexArrays(1)
        else:
            self.vao = None

        self.indexBuffer = None
        self.nindices    = None

        # The loadAtts/unloadAtts methods add
        # to this counter to allow re-entrance
        # (so we don't try to load/unload more
        # than once).
        self.attsLoaded = 0

        log.debug('%s.init(%s)', type(self).__name__, id(self))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('%s.del(%s)', type(self).__name__, id(self))


    def ready(self):
        """Checks whether every uniform and attribute has been given a value.
        Returns a tuple containing:
          - ``True`` if every uniform/attribute has been set, ``False``
            otherwise.
          - A list of the names of all unset uniforms/attributes.
        """
        unset = []
        for n, v in self.hasValue.items():
            if not v:
                unset.append(n)
        return len(unset) == 0, unset


    @contextlib.contextmanager
    def loaded(self):
        """Context manager which calls :meth:`load`, yields, then
        calls :meth:`unload`.
        """
        self.load()
        try:
            yield
        finally:
            self.unload()


    @contextlib.contextmanager
    def loadedAtts(self):
        """Context manager which calls :meth:`loadAtts`, yields, then
        calls :meth:`unloadAtts`.

        This is called automatically by :meth:`draw`, so there is no need
        to explicitly call it.
        """
        self.loadAtts()
        try:
            yield
        finally:
            self.unloadAtts()


    def load(self):
        """Loads this ``GLSLShader`` into the GL state. """
        gl.glUseProgram(self.program)


    def loadAtts(self):
        """Binds all of the shader program ``attribute`` variables - you
        must set the data for each attribute via :meth:`setAtt` before
        calling this method.

        This is called automatically by :meth:`draw`, so there is no need
        to explicitly call it.

        Attributes may be set before or after this method is called.
        """
        self.attsLoaded += 1
        if self.attsLoaded > 1:
            return
        if self.vao is not None:
            gl.glBindVertexArray(self.vao)
        for att in self.attributes:

            aPos           = self.positions[      att]
            aType          = self.types[          att]
            aBuf           = self.buffers[        att]
            aDivisor       = self.attDivisors.get(att)
            glType, glSize = GLSL_ATTRIBUTE_TYPES[aType]

            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, aBuf)
            gl.glEnableVertexAttribArray(aPos)
            gl.glVertexAttribPointer(    aPos,
                                         glSize,
                                         glType,
                                         gl.GL_FALSE,
                                         0,
                                         None)

            if aDivisor is not None:
                glexts.glVertexAttribDivisor(aPos, aDivisor)

        if self.indexBuffer is not None:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.indexBuffer)


    def unloadAtts(self):
        """Disables all vertex attributes, and unbinds associated vertex buffers.
        This is called automatically by :meth:`draw`, so there is no need
        to explicitly call it.
        """
        self.attsLoaded -= 1
        if self.attsLoaded > 0:
            return
        for att in self.attributes:
            gl.glDisableVertexAttribArray(self.positions[att])

            pos     = self.positions[      att]
            divisor = self.attDivisors.get(att)

            if divisor is not None:
                glexts.glVertexAttribDivisor(pos, 0)

        if self.indexBuffer is not None:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
        if self.vao is not None:
            gl.glBindVertexArray(0)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)


    def unload(self):
        """Unloads the GL shader program. """
        gl.glUseProgram(0)


    def destroy(self):
        """Deletes all GL resources managed by this ``GLSLShader``. """
        gl.glDeleteProgram(self.program)
        if self.vao is not None:
            gl.glDeleteVertexArrays(1, self.vao)
        if self.indexBuffer is not None:
            gl.glDeleteBuffers(1, self.indexBuffer)
        for buf in self.buffers.values():
            gl.glDeleteBuffers(1, gltypes.GLuint(buf))

        self.program     = None
        self.vao         = None
        self.indexBuffer = None
        self.buffers     = None


    @memoize.Instanceify(memoize.skipUnchanged)
    def set(self, name, value, size=None):
        """Sets the value for the specified GLSL ``uniform`` variable.

        The ``GLSLShader`` keeps a copy of the value of every uniform, to
        avoid unnecessary GL calls.


        .. note:: This method is decorated by the
                  :func:`.memoize.skipUnchanged` decorator, which returns
                  ``True`` if the value was changed, ``False`` otherwise.
        """

        vPos  = self.positions[name]
        vType = self.types[    name]
        vSize = self.sizes[    name]

        if size is None:
            size = 1

        setfunc = getattr(self, '_uniform_{}'.format(vType), None)

        if size > vSize:
            raise RuntimeError('Specified size ({}) is greater than '
                               'uniform size {} ({})'.format(
                                   size, name, vSize))

        if setfunc is None:
            raise RuntimeError('Unsupported shader program '
                               'type: {}'.format(vType))

        log.debug('Setting shader variable: %s(%s)[%s] = %s',
            vType, size, name, value)

        setfunc(vPos, value, size)
        self.hasValue[name] = True


    def setAtt(self, name, value, divisor=None):
        """Sets the value for the specified GLSL ``attribute`` variable.

        :arg divisor: If specified, this value is used as a divisor for this
                      attribute via the ``glVetexAttribDivisor`` function.

        .. note:: If a ``divisor`` is specified, the OpenGL
                  ``ARB_instanced_arrays`` extension must be
                  available.
        """

        aType    = self.types[  name]
        aBuf     = self.buffers[name]

        castfunc = getattr(self, '_attribute_{}'.format(aType), None)

        if castfunc is None:
            raise RuntimeError('Unsupported shader program '
                               'type: {}'.format(aType))

        value = castfunc(value)

        log.debug('Setting shader attribute: %s(%s): %s',
            aType, name, value.shape)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, aBuf)
        gl.glBufferData(gl.GL_ARRAY_BUFFER,
                        value.nbytes,
                        value,
                        gl.GL_STATIC_DRAW)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

        if divisor is not None:
            self.attDivisors[name] = divisor

        self.hasValue[name] = True


    def setIndices(self, indices):
        """If an index array is to be used by this ``GLSLShader``, the index
        array may be set via this method.
        """

        if indices is None:
            if self.indexBuffer is not None:
                gl.glDeleteBuffers(1, self.indexBuffer)
            self.indexBuffer = None
            self.nindices    = None
            return

        if self.indexBuffer is None:
            self.indexBuffer = gl.glGenBuffers(1)

        self.nindices = indices.size

        indices = np.asarray(indices, dtype=np.uint32)

        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER,
                        self.indexBuffer)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER,
                        indices.nbytes,
                        indices,
                        gl.GL_STATIC_DRAW)

        # Don't unbind if loadAtts is active.
        # This allows setIndices to be called
        # either before or after loadAtts.
        if not self.attsLoaded:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)


    def draw(self, prim, *args):
        """Submits a GL draw call for the specified primitive type.
        If vertex indices have been provided via the :meth:`setIndices` method,
        ``glDrawElements`` is used, and all other arguments are ignored.
        Otherwise, ``glDrawArrays`` is used, and is passed all other arguments.
        """

        with self.loadedAtts():
            if self.indexBuffer is not None:
                gl.glDrawElements(prim,
                                  self.nindices,
                                  gl.GL_UNSIGNED_INT,
                                  None)
            else:
                gl.glDrawArrays(prim, *args)


    def __getPositions(self, shaders, attributes, uniforms):
        """Gets the position indices for all shader attributes (vertex inputs),
        and uniforms (constant inputs) for the given shader programs.

        :arg shaders:    Reference to the compiled shader program.
        :arg attributes: List of attributes required by the shader.
        :arg uniforms:   List of uniforms required by the shader.

        :returns:  A dictionary of ``{name : position}`` mappings.
        """

        import OpenGL.GL as gl

        shaderVars = {}

        for v in uniforms:
            shaderVars[v] = gl.glGetUniformLocation(shaders, v)

        for v in attributes:
            shaderVars[v] = gl.glGetAttribLocation(shaders, v)

        return shaderVars


    def __compile(self, vertShaderSrc, fragShaderSrc, geomShaderSrc=None):
        """Compiles and links the OpenGL GLSL vertex, fragment, and optionally
        geometry shader programs, and returns a reference to the resulting
        program. Raises an error if compilation/linking fails.

        .. note:: I'm explicitly not using the PyOpenGL
                  :func:`OpenGL.GL.shaders.compileProgram` function, because
                  it attempts to validate the program after compilation, which
                  fails due to texture data not being bound at the time of
                  validation.
        """

        program = gl.glCreateProgram()
        srcs    = [(vertShaderSrc, gl.GL_VERTEX_SHADER),
                   (fragShaderSrc, gl.GL_FRAGMENT_SHADER),
                   (geomShaderSrc, gl.GL_GEOMETRY_SHADER)]

        for src, srcType in srcs:
            if src is None:
                continue

            shader = gl.glCreateShader(srcType)
            gl.glShaderSource(shader, src)
            gl.glCompileShader(shader)
            result = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
            if result != gl.GL_TRUE:
                raise RuntimeError(
                    '{}: {}'.format(srcType, gl.glGetShaderInfoLog(shader)))
            gl.glAttachShader(program, shader)
            gl.glDeleteShader(shader)

        gl.glLinkProgram(program)
        linkResult = gl.glGetProgramiv(program, gl.GL_LINK_STATUS)

        if linkResult != gl.GL_TRUE:
            raise RuntimeError(gl.glGetProgramInfoLog(program))

        return program


    def _attribute_bool(self, val):
        return np.array(val, dtype=np.bool, copy=False)


    def _attribute_int(self, val):
        return np.array(val, dtype=np.int32, copy=False)


    def _attribute_float(self, val):
        return np.array(val, dtype=np.float32, copy=False)


    def _attribute_vec2(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _attribute_vec3(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _attribute_vec4(self, val):
        return np.array(val, dtype=np.float32, copy=False).ravel('C')


    def _uniform_bool(self, pos, val, size):
        val = np.array(val, dtype=np.int32)
        gl.glUniform1iv(pos, size, val)


    def _uniform_int(self, pos, val, size):
        val = np.array(val, dtype=np.int32)
        gl.glUniform1iv(pos, size, val)


    def _uniform_float(self, pos, val, size):
        val = np.array(val, dtype=np.float32)
        gl.glUniform1fv(pos, size, val)


    def _uniform_vec2(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform2fv(pos, size, val)


    def _uniform_vec3(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform3fv(pos, size, val)


    def _uniform_vec4(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniform4fv(pos, size, val)


    def _uniform_mat2(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix2fv(pos, size, True, val)


    def _uniform_mat3(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix3fv(pos, size, True, val)


    def _uniform_mat4(self, pos, val, size):
        val = np.array(val, dtype=np.float32, copy=False).ravel('C')
        gl.glUniformMatrix4fv(pos, 1, True, val)


    def _uniform_sampler1D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)


    def _uniform_sampler2D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)


    def _uniform_sampler3D(self, pos, val, size):
        val = np.array(val, dtype=np.int32, copy=False).ravel('C')
        gl.glUniform1iv(pos, size, val)



#!/usr/bin/env python
#
# viewpanel.py - Superclass for all FSLeyes view panels.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ViewPanel` class, which is the base-class
for all of the *FSLeyes view* panels. See the :mod:`fsleyes` package
documentation for more details.
"""


import              logging
import itertools as it

import                                  wx
import wx.lib.agw.aui                as aui
import wx.lib.agw.aui.framemanager   as auifm

import fsl.utils.notifier            as notifier
import fsleyes_widgets               as fwidgets

import fsleyes.panel                 as fslpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.profiles              as profiles
import fsleyes.strings               as strings
import fsleyes.actions               as actions


log = logging.getLogger(__name__)


class ViewPanel(fslpanel.FSLeyesPanel):
    """The ``ViewPanel`` class is the base-class for all *FSLeyes views*.

    A ``ViewPanel`` displays some sort of view of the overlays in an
    :class:`.OverlayList`. The settings for a ``ViewPanel`` are defined by a
    :class:`.DisplayContext` instance.


    **Panels and controls**


    A ``ViewPanel`` class uses a ``wx.lib.agw.aui.AuiManager`` to lay out its
    children. A ``ViewPanel`` has one central panel, which contains the
    primary view; and may have one or more secondary panels, which contain
    *controls* - see the :mod:`.controls` package. The centre panel can be set
    via the :meth:`centrePanel` property, and secondary panels can be
    added/removed to/from with the :meth:`togglePanel` method. The current
    state of a secondary panel (i.e. whether one is open or not) can be
    queried with the :meth:`isPanelOpen` method, and existing secondary panels
    can be accessed via the :meth:`getPanel` method.  Secondary panels must be
    derived from either the :class:`.ControlPanel` or :class:`.ControlToolBar`
    base-classes.


    **Profiles**


    The logic which defines the way that a user interacts with a view panel
    is defined by a  :class:`.Profile` object, which contains mouse and
    keyboard event handlers for reacting to user input.

    Sub-classes must call the :meth:`initProfile` method to initialise their
    default profile. Other profiles may be associated with a particular
    control panel - these profiles can be temporarily activated and deactivated
    when the control panel is toggled via the :meth:`togglePanel` method.

    The currently active interaction profile can be accessed with the
    :meth:`currentProfile` method. See the :mod:`.profiles` package for
    more information on interaction profiles.


    **Programming interface**


    The following methods are available on a ``Viewpanel`` for programmatically
    controlling its display and layout:


    .. autosummary::
       :nosignatures:

       togglePanel
       isPanelOpen
       getPanel
       getPanels
       getTools
       removeFromFrame
       removeAllPanels
       getPanelInfo
       auiManager
    """


    def controlOptions(self, cpType):
        """May be overridden by sub-classes. Given a control panel type,
        may return a dictionary containing arguments to be passed to
        the  ``__init__`` method when the control panel is created.
        """
        return None


    @staticmethod
    def title():
        """May be overridden by sub-classes. Returns a title for this
        ``ViewPanel``, to be used in menus and window title bars.
        """
        return None


    @staticmethod
    def controlOrder():
        """May be overridden by sub-classes. Returns a list of names of
        control panel types, specifying a suggested order for the
        settings menu for views of this type.
        """
        return None


    @staticmethod
    def toolOrder():
        """May be overridden by sub-classes. Returns a list of names of
        tools, specifying a suggested order for the corresponding entries
        in the FSLeyes tools menu. Note that the ordering of tools returned
        by the :meth:`getTools` method is honoured - the ordering returned
        by *this* method relates to tools which are implemented as plugins.
        """
        return None


    @staticmethod
    def defaultLayout():
        """May be overridden by sub-classes. Should return a list of names of
        FSLeyes :class:`.ControlPanel` types which form the default layout for
        this view.
        """
        return None


    @staticmethod
    def displayType():
        """May be overridden by sub-classes. May return a string which
        describes the type of display/view that this ``ViewPanel`` provides.
        This might be used as a hint by :class:`.Display` /
        :class:`.DisplayOpts` instances to tailor their overlay display
        settings to different view/display types.
        """
        return None


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``ViewPanel``.

        :arg parent:         ``wx`` parent object
        :arg overlayList:    The :class:`.OverlayList`
        :arg displayCtx:     A :class:`.DisplayContext` object unique to this
                             ``ViewPanel``
        :arg frame:          The :class:`.FSLeyesFrame`
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        # Currently only two profiles are allowed at
        # any one time - the default profile, passed
        # to initProfiles. and one more profile
        # associated with a control panel.
        #
        # If the default profile is active, and a
        # new control is opened which needs a different
        # profile, that profile is activated.
        #
        # If a non-default profile (#1) is active, and
        # a new control is opened which needs a
        # different profile (#2), all controls which
        # require profile #1 are closed, and profile
        # #2 is activated.
        #
        # This all happens in togglePanel.
        self.__profileManager = profiles.ProfileManager(
            self, overlayList, displayCtx, 2)

        # The __centrePanel attribute stores a reference
        # to the main (centre) panel on this ViewPanel.
        # It is set by sub-class implementations via
        # the centrePanel property.
        #
        # The panels dictionary stores a collection
        # of {type : instance} mappings of active
        # FSLeyes control panels that are contained
        # in this view panel.
        self.__centrePanel = None
        self.__panels      = {}

        # Notifier instance for emitting events.
        # Currently only two events are emitted
        # - one of them is profile changes, and
        # the profilemanager emits these events
        # anyway, so we can just use it. As for
        # the other event (aui layout changes),
        # well, we're being a bit dodgy and
        # emitting these events via the
        # profilemanager.
        self.__events = self.__profileManager

        # See note in FSLeyesFrame about
        # the user of aero docking guides.
        self.__auiMgr = aui.AuiManager(
            self,
            agwFlags=(aui.AUI_MGR_RECTANGLE_HINT          |
                      aui.AUI_MGR_NO_VENETIAN_BLINDS_FADE |
                      aui.AUI_MGR_ALLOW_FLOATING          |
                      aui.AUI_MGR_AERO_DOCKING_GUIDES     |
                      aui.AUI_MGR_LIVE_RESIZE))

        self.__auiMgr.SetDockSizeConstraint(0.5, 0.5)
        self.__auiMgr.Bind(aui.EVT_AUI_PANE_CLOSE,
                           self.__onPaneClose)
        self.__auiMgr.Bind(aui.EVT_AUI_PERSPECTIVE_CHANGED,
                           self.__onPerspectiveChange)

        # A very shitty necessity. When panes are floated,
        # the AuiManager sets the size of the floating frame
        # to the minimum size of the panel, without taking
        # into account the size of its borders/title bar,
        # meaning that the panel size is too small. Here,
        # we're just creating a dummy MiniFrame (from which
        # the AuiFloatingFrame derives), and saving the size
        # of its trimmings for later use in the togglePanel
        # method.
        ff         = wx.MiniFrame(self)
        size       = ff.GetSize().Get()
        clientSize = ff.GetClientSize().Get()
        self.__floatOffset = (size[0] - clientSize[0],
                              size[1] - clientSize[1])
        ff.Destroy()

        self.reloadPlugins()


    def reloadPlugins(self):
        """Called by :meth:`__init__`, and by the :class:`.LoadPluginAction`
        when new plugins are registered.

        This is a bit of a hack, but less hacky than it used to be.  All
        plugin-provided control panels and tools which support this
        ``ViewPanel`` are looked up via the :mod:`.plugins.` module. Then, for
        each control, we create a :class:`.ToggleControlPanelAction`, and add
        it as an attribute on this ``ViewPanel``.

        Similarly, all plugin-provided tools which support this ``ViewPanel``
        are created and added as attributes.

        In both cases, the class name of the control/tool is used as the
        attribute name.

        This is done so that these actions will work with the
        :class:`.ActionProvider` interface, and hence the
        :meth:`.FSLEeyesFrame.populateMenu` method.

        This implementation may change in the future if it becomes problematic
        (e.g. due to naming conflicts).
        """

        import fsleyes.plugins as plugins

        # controls
        for ctrlTitle, ctrlType in plugins.listControls(type(self)).items():
            name = ctrlType.__name__
            if not hasattr(self, name):
                act = actions.ToggleControlPanelAction(
                    self.overlayList,
                    self.displayCtx,
                    self,
                    ctrlType,
                    name=name,
                    title=ctrlTitle)
                setattr(self, name, act)

        # tools
        for toolType in plugins.listTools(type(self)).values():
            name = toolType.__name__
            if not hasattr(self, name):
                act = toolType(self.overlayList, self.displayCtx, self)
                setattr(self, name, act)


    def destroy(self):
        """Removes some property listeners, destroys all child panels,
        destroys the :class:`.ProfileManager`, and ``AuiManager``, and
        calls :meth:`.FSLeyesPanel.destroy`.
        """

        # Order of operations is important. For example,
        # FSLeyesPanel.destroy will result in all Actions
        # associated with this ViewPanel being destroyed,
        # and their destroy routines may interact with
        # this ViewPanel.

        # remove listeners from overlaylist and display context
        lName = 'ViewPanel_{}'.format(self.name)
        self.overlayList.removeListener('overlays',        lName)
        self.displayCtx .removeListener('selectedOverlay', lName)

        fslpanel.FSLeyesPanel.destroy(self)

        # Make sure that any control panels are correctly destroyed
        for panel in self.__panels.values():
            self.__auiMgr.DetachPane(panel)
            panel.destroy()

        # Disable the  ProfileManager
        self.__profileManager.destroy()

        # Un-initialise the AUI manager
        self.__auiMgr.UnInit()

        # The AUI manager does not clear its
        # reference to this panel, so let's
        # do it here.
        self.__auiMgr._frame  = None
        self.__profileManager = None
        self.__auiMgr         = None
        self.__panels         = None
        self.__centrePanel    = None
        self.__events         = None


    @property
    def events(self) -> notifier.Notifier:
        """Return a reference to a :class:`.Notifier` instance which can be
        used to be notified when certain events occur. Currently the following
        events are emitted:

         - ``'profile'``, when the current interaction profile changes.
           Callbacks which are registered with the ``'profile'`` topic will
           be passed a tuple containing the types (:class:`.Profile`
           sub-classes) of the de-registered and newly registered profiles.

         - ``'aui_perspective'``, when the AUI-managed layout changes, e.g.
           sash resizes, control panels added/removed, etc. This event is
           emitted whenever the ``AuiManager`` emits a
           ``EVT_AUI_PERSPECTIVE_CHANGED`` event. It is re-emitted via the
           :class:`.Notifer` interface so that non-wx entities can be
           notified (see e.g. the :class:`.ToggleControlPanelAction`).
        """
        return self.__events


    def initProfile(self, defaultProfile):
        """Must be called by subclasses, after they have initialised all
        of the attributes which may be needed by their associated
        :class:`.Profile` instances.

        :arg defaultProfile: Default profile type
        """
        self.__profileManager.activateProfile(defaultProfile)


    @property
    def currentProfile(self):
        """Returns the :class:`.Profile` instance currently in use. """
        return self.__profileManager.getCurrentProfile()


    @property
    def profileManager(self):
        """Returns a reference to the :class:`.ProfileManager` used by
        this ``ViewPanel``.
        """
        return self.__profileManager


    @property
    def centrePanel(self):
        """Returns the primary (centre) panel on this ``ViewPanel``.
        """
        return self.__centrePanel


    @centrePanel.setter
    def centrePanel(self, panel):
        """Set the primary centre panel for this ``ViewPanel``. This method
        is only intended to be called by sub-classes.
        """

        panel.Reparent(self)
        paneInfo = (aui.AuiPaneInfo()
                    .Name(type(panel).__name__)
                    .CentrePane())

        self.__auiMgr.AddPane(panel, paneInfo)
        self.__auiMgrUpdate()
        self.__centrePanel = panel


    def togglePanel(self, panelType, *args, **kwargs):
        """Add/remove the secondary control panel of the specified type to/from
        this ``ViewPanel``.

        If no keyword argunents are specified, the arguments returned by the
        :meth:`.ControlMixin.defaultLayout` method are returned.

        :arg panelType: Type of the secondary panel.

        :arg args:      All positional arguments are passed to the
                        ``panelType`` constructor.

        :arg floatPane: If ``True``, the secondary panel is initially floated.
                        Defaults to ``False``.

        :arg floatOnly: If ``True``, and ``floatPane=True``, the panel will
                        be permanently floated (i.e. it will not be dockable).

        :arg floatPos:  If provided, and ``floatPane`` is ``True``, specifies
                        the location of the floating panel as ``(w, h)``
                        proportions between 0 and 1, relative to this view
                        panel.

        :arg closeable: If ``False``, and ``floatPane=True``, the panel will
                        not have a close button when it is floated.

        :arg location:  If ``floatPane=False``, the initial dock position of
                        the panel - either ``wx.TOP``, ``wx.BOTTOM``,
                        ``wx.LEFT``, or ``wx.RIGHT. Defaults to ``wx.BOTTOM``.

        :arg title:     Title to give the control. If not provided, it is
                        assumed that a title for ``panelType`` is in
                        :attr:`.strings.titles`.

        :arg kwargs:    All other keyword arguments, are passed to the
                        ``panelType`` constructor.

        :returns:       The newly created control panel.

        .. note::       The ``panelType`` type must be a sub-class of
                        :class:`.ControlPanel` or :class:`.ControlToolBar`,
                        which can be created like so::

                            panel = panelType(parent,
                                              overlayList,
                                              displayCtx,
                                              frame,
                                              *args,
                                              **kwargs)

        .. warning:: Do not define a control (a.k.a. secondary) panel
                     constructor to accept arguments with the names
                     ``floatPane``, ``floatOnly``, ``floatPos``,
                     ``closeable``, or ``location``, as arguments with those
                     names will get eaten by this method before they can be
                     passed to the constructor.
        """

        # Merge in default options for this control,
        # with precedence (highest to lowest):
        #  1. kwargs
        #  2. ViewPanel.controlOptions
        #  3. ControlPanel.defaultLayout
        layout = panelType.defaultLayout()
        cpopts = self.controlOptions(panelType)

        if layout is None: layout = {}
        if cpopts is None: cpopts = {}

        for k, v in it.chain(cpopts.items(), layout.items()):
            if k not in kwargs:
                kwargs[k] = v

        location   = kwargs.pop('location',  None)
        floatPane  = kwargs.pop('floatPane', False)
        floatOnly  = kwargs.pop('floatOnly', False)
        closeable  = kwargs.pop('closeable', True)
        title      = kwargs.pop('title',     None)
        floatPos   = kwargs.pop('floatPos',  (0.5, 0.5))
        profileCls = panelType.profileCls()

        if title is None:
            title = panelType.title()
        if title is None:
            title = strings.titles.get(panelType, panelType.__name__)

        if location not in (None, wx.TOP, wx.BOTTOM, wx.LEFT, wx.RIGHT):
            raise ValueError('Invalid value for location')

        supported = panelType.supportedViews()
        if supported is not None and not isinstance(self, tuple(supported)):
            raise ValueError(
                '{} views are not supported by {} controls'.format(
                    type(self).__name__, panelType.__name__))

        # The panel is already open - close it
        window = self.__panels.get(panelType, None)
        if window is not None:
            self.__onPaneClose(None, window)
            return

        # Otherwise, create a new panel of the specified type.
        # If this control is associated with a custom interaction
        # profile, check that that profile is active, create it
        # if needed, and close down any controls which are not
        # compatible with the new profile.
        profile = self.currentProfile
        if profileCls is not None and not isinstance(profile, profileCls):

            log.debug('New control %s requires interaction profile %s but '
                      'profile %s is active', panelType.__name__,
                      profileCls.__name__, type(profile).__name__)

            # close down incompatible controls
            for ctype, cpanel in list(self.__panels.items()):
                cprofileCls = ctype.profileCls()
                if cprofileCls is not None and \
                   not issubclass(cprofileCls, profileCls):
                    log.debug('Closing down control %s as it is incompatible '
                              'with profile %s', ctype.__name__,
                              profileCls.__name__)
                    self.__onPaneClose(None, cpanel)

            # change profile
            self.__profileManager.activateProfile(profileCls)

        # The PaneInfo Name is the control panel class name -
        # this is used for saving and restoring layouts.
        window = panelType(self,
                           self.overlayList,
                           self.displayCtx,
                           self,
                           *args,
                           **kwargs)
        # save a ref to the control if it has a custom
        # profile so, when it is destroyed, we know
        if profileCls is not None:
            self.__activeControl = window

        paneInfo = self.__layoutNewPanel(window, title, location, floatPane,
                                         floatPos, floatOnly, closeable)
        self.__auiMgr.AddPane(window, paneInfo)
        self.__panels[panelType] = window
        self.__auiMgrUpdate(newPanel=window)
        return window


    def __layoutNewPanel(self,
                         panel,
                         title,
                         location,
                         floatPane,
                         floatPos,
                         floatOnly,
                         closeable):
        """Sub-method of :meth:`togglePanel`. Creates and returns an
        ``AuiPaneInfo`` instance describing the initial layout of a new
        control panel. See ``togglePanel`` for an explanation of the
        arguments.
        """
        paneInfo  = aui.AuiPaneInfo().Name(type(panel).__name__)
        isToolbar = isinstance(panel, ctrlpanel.ControlToolBar)

        if isToolbar:

            # ToolbarPane sets the panel layer to 10
            paneInfo.ToolbarPane()

            if panel.GetOrient() == wx.VERTICAL:
                paneInfo.GripperTop()

            # We are going to put any new toolbars on
            # the top of the panel, below any existing
            # toolbars. This is annoyingly complicated,
            # because the AUI designer(s) decided to
            # give the innermost layer an index of 0.
            #
            # So in order to put a new toolbar at the
            # innermost layer, we need to adjust the
            # layers of all other existing toolbars

            for p in self.__panels.values():
                if isinstance(p, ctrlpanel.ControlToolBar):
                    info = self.__auiMgr.GetPane(p)

                    # This is nasty - the agw.aui.AuiPaneInfo
                    # class doesn't have any publicly documented
                    # methods of querying its current state.
                    # So I'm accessing its undocumented instance
                    # attributes (determined by browsing the
                    # source code)
                    if info.IsDocked() and \
                       info.dock_direction == aui.AUI_DOCK_TOP:
                        info.Layer(info.dock_layer + 1)

            # When the toolbar contents change,
            # update the layout, so that the
            # toolbar's new size is accommodated
            panel.Bind(fsltoolbar.EVT_TOOLBAR_EVENT, self.__auiMgrUpdate)

        paneInfo.Caption(title)

        # Dock the pane at the position specified
        # by the location parameter
        if not floatPane:

            if location is None:
                if isToolbar: location = aui.AUI_DOCK_TOP
                else:         location = aui.AUI_DOCK_BOTTOM

            elif location == wx.TOP:    location = aui.AUI_DOCK_TOP
            elif location == wx.BOTTOM: location = aui.AUI_DOCK_BOTTOM
            elif location == wx.LEFT:   location = aui.AUI_DOCK_LEFT
            elif location == wx.RIGHT:  location = aui.AUI_DOCK_RIGHT

            # Make sure the pane is
            # resizable in case it
            # gets floated later on
            paneInfo.Direction(location) \
                    .Resizable(not isToolbar)

        # Or, for floating panes, centre the
        # floating pane on this ViewPanel
        else:

            selfPos    = self.GetScreenPosition().Get()
            selfSize   = self.GetSize().Get()
            selfCentre = (selfPos[0] + selfSize[0] * floatPos[0],
                          selfPos[1] + selfSize[1] * floatPos[1])

            paneSize = panel.GetBestSize().Get()
            panePos  = (int(round(selfCentre[0] - paneSize[0] * 0.5)),
                        int(round(selfCentre[1] - paneSize[1] * 0.5)))

            paneInfo.Float()                 \
                    .Resizable(True)         \
                    .Dockable(not floatOnly) \
                    .CloseButton(closeable)  \
                    .FloatingPosition(panePos)

        return paneInfo


    def isPanelOpen(self, panelType):
        """Returns ``True`` if a panel of type ``panelType`` is open,
        ``False`` otherwise.
        """
        return self.getPanel(panelType) is not None


    def getPanel(self, panelType):
        """If an instance of ``panelType`` exists, it is returned.
        Otherwise ``None`` is returned.
        """
        if panelType in self.__panels: return self.__panels[panelType]
        else:                          return None


    @actions.action
    def removeAllPanels(self):
        """Remove all control panels from this ``ViewPanel``."""
        for panelType in list(self.__panels.keys()):
            self.togglePanel(panelType)


    @actions.action
    def removeFromFrame(self):
        """Remove this ``ViewPanel`` from the :class:`.FSLeyesFrame`.

        Will raise an error if this ``ViewPanel`` is not in a
        ``FSLeyesFrame``.
        """
        self.frame.removeViewPanel(self)


    def getPanels(self):
        """Returns a list containing all control panels currently shown in this
        ``ViewPanel``.
        """
        return list(self.__panels.values())


    def getPanelInfo(self, panel):
        """Returns the ``AuiPaneInfo`` object which contains information about
        the given control panel.
        """
        return self.__auiMgr.GetPane(panel)


    @property
    def auiManager(self):
        """Returns the ``wx.lib.agw.aui.AuiManager`` object which manages the
        layout of this ``ViewPanel``.
        """
        return self.__auiMgr


    def getTools(self):
        """This method should be overridden by sub-classes (if necessary), and
        should return any ``action`` methods which should be added to the
        :class:`.FSLeyesFrame` *Tools* menu.

        See also the :meth:`.ActionProvider.getActions` method, which can
        also be overridden, and controls the actions which get added to the
        FSLeyes *settings* menu.
        """
        return []


    def __auiMgrUpdate(self, *args, **kwargs):
        """Called whenever a panel is added/removed to/from this ``ViewPanel``.

        Calls the ``Update`` method on the ``AuiManager`` instance that is
        managing this panel.

        :arg newPanel: Must be passed as a keyword argument. When a new panel
                       is added, it should be passed here.
        """

        newPanel = kwargs.pop('newPanel', None)

        # This method makes sure that size hints
        # for all existing and new panels are
        # set on their AuiPaneInfo objects, and
        # then calls AuiManager.Update.

        # We first loop through all panels, and
        # figure out their best sizes. Each entry
        # in this list is a tuple containing:
        #
        #    - Panel
        #    - AuiPaneInfo instance
        #    - Dock direction (None for floating panels)
        #    - Layer number (None for floating panels)
        #    - Minimum size
        bestSizes = []

        for panel in self.__panels.values():

            if isinstance(panel, ctrlpanel.ControlToolBar):
                continue

            pinfo = self.__auiMgr.GetPane(panel)

            # If the panel is floating, use its
            # current size as its 'best' size,
            # as otherwise the AuiManager will
            # immediately resize the panel to
            # its best size.
            if pinfo.IsFloating():
                dockDir  = None
                layer    = None
                bestSize = panel.GetSize().Get()

                # Unless its current size is tiny
                # (which probably means that it has
                # just been added)
                if bestSize[0] <= 20 or \
                   bestSize[1] <= 20:
                    bestSize = panel.GetBestSize().Get()

            else:
                dockDir  = pinfo.dock_direction
                layer    = pinfo.dock_layer
                bestSize = panel.GetBestSize().Get()

            bestSizes.append((panel, pinfo, dockDir, layer, bestSize))

        # Now we loop through one final time, and
        # set all of the necessary size hints on
        # the AuiPaneInfo instances.
        for panel, pinfo, dockDir, layer, bestSize in bestSizes:

            parent = panel.GetParent()

            # When a panel is added/removed from the AuiManager,
            # the position of floating panels seems to get reset
            # to their original position, when they were created.
            # Here, we explicitly set the position of each
            # floating frame, so the AuiManager doesn't move our
            # windows about the place.
            if pinfo.IsFloating() and \
               isinstance(parent, aui.AuiFloatingFrame):
                pinfo.FloatingPosition(parent.GetScreenPosition())

            # See comments in __init__ about
            # this silly 'float offset' thing
            floatSize = (bestSize[0] + self.__floatOffset[0],
                         bestSize[1] + self.__floatOffset[1])

            log.debug('New size for panel {} - '
                      'best: {}, float: {}'.format(
                          type(panel).__name__, bestSize, floatSize))

            pinfo.MinSize(     (1, 1))  \
                 .BestSize(    bestSize) \
                 .FloatingSize(floatSize) \
                 .Resizable(   True)

            # This is a terrible hack which forces
            # the AuiManager to grow a dock when a
            # new panel is added, which is bigger
            # than the existing dock contents.
            if panel is newPanel and not pinfo.IsFloating():
                docks = aui.FindDocks(self.__auiMgr._docks, dockDir, layer)
                for d in docks:
                    d.size = 0

        self.__auiMgr.Update()


    def __onPaneClose(self, ev=None, panel=None):
        """Called when the user closes a control (a.k.a. secondary) panel.
        Calls the
        :class:`.ControlPanel.destroy`/:class:`.ControlToolBar.destroy`
        method on the panel.
        """

        if ev is not None:
            ev.Skip()
            panel = ev.GetPane().window

        # If the user has grouped multiple control panels
        # into a single tabbed notebook, and then closed
        # the entire notebook, the AuiManager will generate
        # a single close event, and will pass us that
        # notebook. So we have to look in the notebook
        # to see which control panels were actually closed.
        if isinstance(panel, wx.lib.agw.aui.AuiNotebook):
            panels = [panel.GetPage(i) for i in range(panel.GetPageCount())]
        else:
            panels = [panel]

        for panel in list(panels):

            # note: in theory, all panels  should be sub-classes
            # of ControlPanel/ControlToolBar. But this check is
            # kept here to support third party scripts which don't
            # honour the FSLeyes plugin rules.
            if isinstance(panel, (ctrlpanel.ControlPanel,
                                  ctrlpanel.ControlToolBar)):

                # WTF AUI. Sometimes this method gets called
                # twice for a panel, the second time with a
                # reference to a wx._wxpyDeadObject; in such
                # situations, the Destroy method call below
                # would result in an exception being raised.
                if self.__panels.pop(type(panel), None) is None:
                    panels.remove(panel)

                # calling ControlPanel.destroy()
                # here -  wx.Destroy is done below
                else:
                    log.debug('Panel closed: %s', type(panel).__name__)
                    panel.destroy()

        # Destroy all the panels
        for panel in panels:

            # Even when the user closes a pane,
            # AUI does not detach said pane -
            # we have to do it manually
            self.__auiMgr.DetachPane(panel)
            wx.CallAfter(panel.Destroy)

        # Update interaction profile. We do not
        # consider multiple tabbed controls here.
        # If the closed control is associated with
        # an interaction profile, if no other other
        # open panels rely on the same destroy the
        # profile and restore the default profile
        #
        # We assume that, if a panel which requires
        # a custom profile was open, that profile
        # was active.
        #
        # See WTF AUI comment above for reason for
        # len(panels) guard
        if len(panels) > 0:
            profileCls   = panels[0].profileCls()
            closeProfile = True
            if profileCls is not None:
                for ctype in self.__panels:
                    cprofileCls = ctype.profileCls()
                    if cprofileCls is not None and \
                       issubclass(cprofileCls, profileCls):
                        closeProfile = False
                if closeProfile:
                    log.debug('Panel %s uses a custom interaction profile %s - '
                              'deactivating it and restoring default profile',
                              type(panels[0]).__name__, profileCls.__name__)
                    self.__profileManager.deactivateProfile()

        # Update the view panel layout
        wx.CallAfter(self.__auiMgrUpdate)


    def __onPerspectiveChange(self, ev):
        """Called on ``EVT_AUI_PERSPECTIVE_CHANGED`` events. Re-emits the
        event via the :meth:`events` notifier, with topic ``'aui_perspective'``.
        This is performed for the benefit of non-wx entities which need to
        know about layout changes.
        """
        self.events.notify(topic='aui_perspective')



#!/usr/bin/env python
#
# orthoannotateprofile.py - The OrthoAnnotateProfile
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoAnnotateProfile` class, an
interaction :class:`.Profile` for :class:`.OrthoPanel` views.
"""

import copy

import numpy as np
import          wx

import fsleyes_widgets.utils.status      as status
import fsleyes_widgets.dialog            as fsldlg
import fsleyes.strings                   as strings
import fsleyes.gl.annotations            as annotations
import fsleyes.profiles.orthoviewprofile as orthoviewprofile


class OrthoAnnotateProfile(orthoviewprofile.OrthoViewProfile):
    """The ``OrthoAnnotateProfile`` class is a :class:`.Profile` for the
    :class:`.OrthoPanel` class, which allows the user to annotate the
    canvases of an ``OrthoPanel`` with simple shapes and text.
    """


    colour = copy.copy(annotations.AnnotationObject.colour)
    """Initial colour to give all annotations. """


    lineWidth =  copy.copy(annotations.AnnotationObject.lineWidth)
    """Initial width to give line-based annotations. """


    fontSize = copy.copy(annotations.TextAnnotation.fontSize)
    """Initial font size to give text annotations. """


    filled = copy.copy(annotations.Rect.filled)
    """Whether ellipses/rectangles are filled in or not."""


    border = copy.copy(annotations.Rect.border)
    """Whether ellipses/rectangles are drawn with a border or not."""


    honourZLimits = copy.copy(annotations.AnnotationObject.honourZLimits)
    """Whether annotations are drawn when outside their Z limits."""


    alpha = copy.copy(annotations.AnnotationObject.alpha)
    """Opacity."""


    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``OrthoAnnotateProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        return {
            ('line',     wx.WXK_SHIFT)                  : 'nav',
            ('line',     wx.WXK_CONTROL)                : 'move',
            ('line',     wx.WXK_ALT)                    : 'pan',
            ('line',    (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice',
            ('point',    wx.WXK_SHIFT)                  : 'nav',
            ('point',    wx.WXK_CONTROL)                : 'move',
            ('point',    wx.WXK_ALT)                    : 'pan',
            ('point',   (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice',
            ('rect',     wx.WXK_SHIFT)                  : 'nav',
            ('rect',     wx.WXK_CONTROL)                : 'move',
            ('rect',     wx.WXK_ALT)                    : 'pan',
            ('rect',    (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice',
            ('text',     wx.WXK_SHIFT)                  : 'nav',
            ('text',     wx.WXK_CONTROL)                : 'move',
            ('text',     wx.WXK_ALT)                    : 'pan',
            ('text',    (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice',
            ('ellipse',  wx.WXK_SHIFT)                  : 'nav',
            ('ellipse',  wx.WXK_CONTROL)                : 'move',
            ('ellipse',  wx.WXK_ALT)                    : 'pan',
            ('ellipse', (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice',
            ('arrow',    wx.WXK_SHIFT)                  : 'nav',
            ('arrow',    wx.WXK_CONTROL)                : 'move',
            ('arrow',    wx.WXK_ALT)                    : 'pan',
            ('arrow',   (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice'}


    @staticmethod
    def altHandlers():
        """Returns the alternate handlers map, which allows event handlers
        defined in one mode to be re-used whilst in another mode.
        """
        return {
            ('line',    'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('point',   'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('rect',    'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('text',    'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('arrow',   'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),
            ('ellipse', 'MiddleMouseDrag') : ('pan', 'LeftMouseDrag'),

            ('move',    'MouseWheel') : ('zoom', 'MouseWheel'),

            # Right mouse click/drag allows
            # annotations to be moved
            ('nav',     'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('line',    'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('point',   'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('rect',    'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('text',    'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('arrow',   'RightMouseDown') : ('move', 'LeftMouseDown'),
            ('ellipse', 'RightMouseDown') : ('move', 'LeftMouseDown'),

            ('nav',     'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('line',    'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('point',   'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('rect',    'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('text',    'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('arrow',   'RightMouseDrag') : ('move', 'LeftMouseDrag'),
            ('ellipse', 'RightMouseDrag') : ('move', 'LeftMouseDrag'),

            ('nav',     'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('line',    'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('point',   'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('rect',    'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('text',    'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('arrow',   'RightMouseUp') : ('move', 'LeftMouseUp'),
            ('ellipse', 'RightMouseUp') : ('move', 'LeftMouseUp')}


    def __init__(self, viewPanel, overlayList, displayCtx):
        """Create an ``OrthoAnnotateProfile``.

        :arg viewPanel:    An :class:`.OrthoPanel` instance.
        :arg overlayList:  The :class:`.OverlayList` instance.
        :arg displayCtx:   The :class:`.DisplayContext` instance.
        """

        orthoviewprofile.OrthoViewProfile.__init__(
            self,
            viewPanel,
            overlayList,
            displayCtx,
            ['line', 'arrow', 'point', 'rect', 'text', 'ellipse', 'move'])
        self.mode = 'nav'

        # Used to store a reference to an annotation
        # and previous mouse location during mouse
        # drags.
        self.__dragging = None
        self.__lastPos  = None


    def __initialSettings(self, canvas, canvasPos):
        """Returns a dictionary containing some initial settings with which all
        new annotations are created.
        """
        opts = canvas.opts
        zpos = canvasPos[opts.zax]
        return {
            'colour'        : self.colour,
            'lineWidth'     : self.lineWidth,
            'fontSize'      : self.fontSize,
            'filled'        : self.filled,
            'border'        : self.border,
            'alpha'         : self.alpha,
            'honourZLimits' : self.honourZLimits,
            'zmin'          : np.floor(zpos),
            'zmax'          : np.ceil( zpos),
            'hold'          : True,
            'fixed'         : False
        }


    def __displaySize(self, size, squared):
        """Display the given size (length or area) in the
        :class:`.FSLeyesFrame` status bar.

        :arg size:    Size to display
        :arg squared: If ``True``, ^2 is shown after the size value (use if
                      the size is an area).
        """
        displayCtx = self.displayCtx
        opts       = displayCtx.getOpts(displayCtx.getSelectedOverlay())
        refimage   = opts.referenceImage

        if refimage is not None:
            units  = refimage.xyzUnits
            units  = strings.nifti.get(('xyz_unit', units), '(unknown units)')

            if squared:
                units = f'{units}\u00B2'
            size = f'{size:.2f} {units}'
        else:
            size = f'{size:.2f}'

        status.update(size)


    def _moveModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """If the mouse lands on an annotation, save a reference to it
        so it can be moved on mouse drag.
        """
        opts  = canvas.opts
        annot = canvas.getAnnotations()
        pos   = canvasPos[opts.xax], canvasPos[opts.yax]

        for obj in annot.annotations:
            try:
                if obj.hit(*pos):
                    self.__dragging = obj
                    self.__lastPos  = pos
                    break
            except NotImplementedError:
                pass


    def _moveModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Move the annotation that was clicked on. """

        obj     = self.__dragging
        lastPos = self.__lastPos

        if obj is None:
            return

        opts   = canvas.opts
        pos    = (canvasPos[opts.xax], canvasPos[opts.yax])
        offset = (pos[0] - lastPos[0], pos[1] - lastPos[1])

        try:
            obj.move(*offset)
            self.__lastPos = pos
        except NotImplementedError:
            pass

        canvas.Refresh()


    def _moveModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clears the reference to the annotation that was being moved. """
        self.__dragging = None
        self.__lastPos  = None


    def _lineModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Adds a new line annotation."""
        opts            = canvas.opts
        annot           = canvas.getAnnotations()
        x, y            = (canvasPos[opts.xax], canvasPos[opts.yax])
        settings        = self.__initialSettings(canvas, canvasPos)
        self.__dragging = annot.line(x, y, x, y, **settings)


    def _lineModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Adjust the line end point so it tracks the mouse location."""
        opts    = canvas.opts
        line    = self.__dragging
        line.x2 = canvasPos[opts.xax]
        line.y2 = canvasPos[opts.yax]

        # display line length in the
        # FSLeyesFrame status bar
        xy1    = np.array([line.x1, line.y1])
        xy2    = np.array([line.x2, line.y2])
        length = np.sqrt(np.sum((xy1 - xy2) ** 2))
        self.__displaySize(length, False)
        canvas.Refresh()


    def _lineModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clear a reference to the newly created line. If the mouse hasn't
        moved since mouse down, the line is deleted.
        """
        line            = self.__dragging
        annot           = canvas.getAnnotations()
        self.__dragging = None

        if (line.x1 == line.x2 and line.y1 == line.y2):
            annot.dequeue(line, hold=True)

        canvas.Refresh()


    def _arrowModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Adds a new arrow annotation."""
        opts            = canvas.opts
        annot           = canvas.getAnnotations()
        x, y            = (canvasPos[opts.xax], canvasPos[opts.yax])
        settings        = self.__initialSettings(canvas, canvasPos)
        self.__dragging = annot.arrow(x, y, x, y, **settings)


    def _arrowModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Adjust the arrow end point so it tracks the mouse location."""
        opts     = canvas.opts
        arrow    = self.__dragging
        arrow.x2 = canvasPos[opts.xax]
        arrow.y2 = canvasPos[opts.yax]

        # display arrow length in the
        # FSLeyesFrame status bar
        xy1    = np.array([arrow.x1, arrow.y1])
        xy2    = np.array([arrow.x2, arrow.y2])
        length = np.sqrt(np.sum((xy1 - xy2) ** 2))
        self.__displaySize(length, False)
        canvas.Refresh()


    def _arrowModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clear a reference to the newly created arrow. If the mouse hasn't
        moved since mouse down, the arrow is deleted.
        """
        arrow            = self.__dragging
        annot           = canvas.getAnnotations()
        self.__dragging = None

        if (arrow.x1 == arrow.x2 and arrow.y1 == arrow.y2):
            annot.dequeue(arrow, hold=True)

        canvas.Refresh()


    def _pointModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Creates a new point annotation. """
        opts            = canvas.opts
        annot           = canvas.getAnnotations()
        x, y            = (canvasPos[opts.xax], canvasPos[opts.yax])
        settings        = self.__initialSettings(canvas, canvasPos)
        self.__dragging = annot.point(x, y, **settings)
        canvas.Refresh()


    def _pointModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Changes the location of the point annotation to track the mouse drag
        location.
        """
        opts              = canvas.opts
        self.__dragging.x = canvasPos[opts.xax]
        self.__dragging.y = canvasPos[opts.yax]
        canvas.Refresh()


    def _pointModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clear a reference to the newly created point annotation. """
        self.__dragging = None


    def _textModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Show a dialog prompting the user for some text, then creates a new
        text annotation.
        """
        opts     = canvas.opts
        annot    = canvas.getAnnotations()
        settings = self.__initialSettings(canvas, canvasPos)
        x, y     = (canvasPos[opts.xax], canvasPos[opts.yax])
        msg      = strings.messages[self, 'TextAnnotation']
        dlg      = fsldlg.TextEditDialog(self.viewPanel,
                                         message=msg,
                                         style=fsldlg.TED_OK_CANCEL)

        if dlg.ShowModal() == wx.ID_OK:
            annot.text(dlg.GetText(), x, y, coordinates='display', **settings)
            canvas.Refresh()


    def _rectModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Create a new rectangle annotation.
        """
        opts            = canvas.opts
        annot           = canvas.getAnnotations()
        x, y            = (canvasPos[opts.xax], canvasPos[opts.yax])
        settings        = self.__initialSettings(canvas, canvasPos)
        self.__dragging = annot.rect(x, y, 0, 0, **settings)


    def _rectModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Adjust the size of the rectangle with the mouse drag. """
        opts   = canvas.opts
        rect   = self.__dragging
        rect.w = canvasPos[opts.xax] - rect.x
        rect.h = canvasPos[opts.yax] - rect.y

        # display rect area in status bar
        self.__displaySize(np.abs(rect.w * rect.h), True)

        canvas.Refresh()


    def _rectModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clear the reference to the new rectangle annotation. If the
        rectangle has no area (the user clicked without dragging), the
        rectangle is deleted.
        """
        rect            = self.__dragging
        annot           = canvas.getAnnotations()
        self.__dragging = None

        if rect.w == 0 or rect.h == 0:
            annot.dequeue(rect, hold=True)

        canvas.Refresh()


    def _ellipseModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Create a new ellipse annotation. """
        opts            = canvas.opts
        annot           = canvas.getAnnotations()
        x, y            = (canvasPos[opts.xax], canvasPos[opts.yax])
        settings        = self.__initialSettings(canvas, canvasPos)
        self.__dragging = annot.ellipse(x, y, 0, 0, **settings)


    def _ellipseModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Adjust the ellipse radius with the mouse drag. """
        opts      = canvas.opts
        ellipse   = self.__dragging
        p1        = np.array((ellipse.x, ellipse.y))
        p2        = np.array((canvasPos[opts.xax], canvasPos[opts.yax]))
        ellipse.w = np.abs(p1[0] - p2[0])
        ellipse.h = np.abs(p1[1] - p2[1])

        # display ellipse area in status bar
        self.__displaySize(np.pi * ellipse.w * ellipse.h, True)
        canvas.Refresh()


    def _ellipseModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Clear the reference to the new ellipse annotation. If the ellipse
        has no area (the user clicked without dragging), the ellipse is deleted.
        """
        ellipse         = self.__dragging
        annot           = canvas.getAnnotations()
        self.__dragging = None

        if (ellipse.w == 0) or (ellipse.h == 0):
            annot.dequeue(ellipse, hold=True)

        canvas.Refresh()



#!/usr/bin/env python
#
# toolbar.py - FSLeyes toolbars
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`FSLeyesToolBar` class, the base class
for all toolbars in *FSLeyes*.
"""

import logging

import wx
import wx.lib.newevent as wxevent

import numpy as np

import fsleyes.panel as fslpanel
import fsleyes.icons as icons


log = logging.getLogger(__name__)


class FSLeyesToolBar(fslpanel.FSLeyesPanel):
    """Base class for all *FSLeyes* toolbars.

    The ``FSLeyesToolBar`` is a regular :class:`wx.PyPanel` which to which a
    group of *tools* can be added, where a tool may be any ``wx`` control.

    See also the :class:`.ControlToolBar`, which is the true base-class for
    all toolbars that are added to FSLeyes view panels.

    Tools can be added to a ``FSLeyesToolBar`` with the following methods:

      .. autosummary::
         :nosignatures:

         AddTool
         InsertTool
         InsertTools
         SetTools
         MakeLabelledTool


    When the horizontal size of a ``FSLeyesToolBar`` becomes too small to
    display all of its tools, the toolbar is compressed: some tools are
    hidden, and buttons are displayed on each end of the toolbar, allowing the
    user to scroll through the toolbar, to access the hidden tools. The user
    may also use the mouse wheel to scroll through the toolbar.

    A collapsed ``FSLeyesToolBar`` looks something like this:

    .. image:: images/fsleyestoolbar.png
       :scale: 50%
       :align: center
    """


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 viewPanel,
                 height=32,
                 orient=wx.HORIZONTAL,
                 *args,
                 **kwargs):
        """Create a ``FSLeyesToolBar``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList`, containing all overlays
                          being displayed.

        :arg displayCtx:  A :class:`.DisplayContext`, which defines how the
                          overlays are to be displayed.

        :arg viewPanel:   The :class:`.ViewPanel` that owns this toolbar.

        :arg height:      Desired toolbar height in pixels. This value is used
                          to look up appropriately sized left/right arrow
                          icons.

        :arg actionz:     A dictionary of actions passed through to the
                          :meth:`.ActionProvider.__init__`.

        All other arguments are passed through to
        :meth:`.FSLeyesPanel.__init__`.
        """

        if orient not in (wx.HORIZONTAL, wx.VERTICAL):
            raise ValueError('Invalid orientation: {}'.format(orient))

        fslpanel.FSLeyesPanel.__init__(self,
                                       parent,
                                       overlayList,
                                       displayCtx,
                                       viewPanel.frame,
                                       *args,
                                       **kwargs)

        self.__tools         = []
        self.__visibleOffset = 0
        self.__numVisible    = 0
        self.__height        = height
        self.__orient        = orient

        font = self.GetFont()
        self.SetFont(font.Smaller())

        style = wx.BU_EXACTFIT | wx.BU_NOTEXT

        if orient == wx.HORIZONTAL:
            lBmp = icons.loadBitmap('thinLeftArrow{}' .format(height))
            rBmp = icons.loadBitmap('thinRightArrow{}'.format(height))
        else:
            lBmp = icons.loadBitmap('thinUpArrow{}'  .format(height))
            rBmp = icons.loadBitmap('thinDownArrow{}'.format(height))

        self.__leftButton  = wx.Button(self, style=style)
        self.__rightButton = wx.Button(self, style=style)

        self.__leftButton .SetBitmap(lBmp)
        self.__rightButton.SetBitmap(rBmp)

        for btn in [self.__leftButton, self.__rightButton]:
            size = btn.GetBestSize()
            btn.SetMinSize(size)

        self.__sizer = wx.BoxSizer(orient)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__leftButton,  flag=wx.EXPAND)
        self.__sizer.Add((0, 0),             flag=wx.EXPAND, proportion=1)
        self.__sizer.Add(self.__rightButton, flag=wx.EXPAND)

        self.__leftButton .Bind(wx.EVT_BUTTON,     self.__onLeftButton)
        self.__rightButton.Bind(wx.EVT_BUTTON,     self.__onRightButton)
        self              .Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)
        self              .Bind(wx.EVT_SIZE,       self.__drawToolBar)



    def GetOrient(self):
        """Returns the orientation of this ``FSLeyesToolBar``, either
        ``wx.HORIZONTAL`` or ``wx.VERTICAL``.
        """
        return self.__orient


    def MakeLabelledTool(self,
                         tool,
                         labelText,
                         labelSide=wx.TOP,
                         expand=False):
        """Creates a panel containing the given tool, and a label for the
        tool. The panel is returned, but is not added to this
        ``FSLeyesToolBar`` - you will have to do that yourself, e.g.::

            labelledTool = toolbar.MakeLabelledTool(tool, 'Label', wx.BOTTOM)
            toolbar.AddTool(labelledTool)

        :arg tool:      A :mod:`wx` control.

        :arg labelText: A label for the tool.

        :arg labelSide: Which side of the tool to put the label - ``wx.TOP``,
                        ``wx.BOTTOM``, ``wx.LEFT``, or ``wx.RIGHT``.

        :arg expand:    Defaults to ``False``. If ``True``, the widget and
                        label will be set up so they expand to fit all
                        available space
        """

        if   labelSide in (wx.TOP,  wx.BOTTOM): orient = wx.VERTICAL
        elif labelSide in (wx.LEFT, wx.RIGHT):  orient = wx.HORIZONTAL

        oldParent = tool.GetParent()
        panel     = wx.Panel(oldParent)
        sizer     = wx.BoxSizer(orient)

        panel.SetSizer(sizer)
        tool.Reparent(panel)

        label = wx.StaticText(panel, style=wx.ALIGN_CENTRE_HORIZONTAL)
        label.SetLabel(labelText)

        if expand:
            sizerArgs = {
                'flag'       : wx.EXPAND,
                'proportion' : 1
            }
        else:
            sizerArgs = {
                'flag' : wx.ALIGN_CENTRE,
            }

        if labelSide in (wx.TOP, wx.LEFT):
            sizer.Add(label, **sizerArgs)
            sizer.Add(tool,  **sizerArgs)
        else:
            sizer.Add(tool,  **sizerArgs)
            sizer.Add(label, **sizerArgs)

        return panel


    def Enable(self, *args, **kwargs):
        """Enables/disables all tools in this ``FSLeyesToolBar``.

        :arg args:   Passed to the ``Enable`` method of each tool.
        :arg kwargs: Passed to the ``Enable`` method of each tool.
        """
        super(FSLeyesToolBar, self).Enable(*args, **kwargs)
        for t in self.__tools:
            t.Enable(*args, **kwargs)


    def GetTools(self):
        """Returns a list containing all tools in this ``FSLeyesToolBar``. """
        return self.__tools[:]


    def GetToolCount(self):
        """Returns the number of tools in this ``FSLeyesToolBar``. """
        return len(self.__tools)


    def AddDivider(self):
        """Adds a :class:`.ToolBarDivider` to the end of the toolbar. """
        self.InsertDivider()


    def InsertDivider(self, index=None):
        """Inserts a :class:`.ToolBarDivider` into the toolbar at the
        specified ``index``.
        """

        if   self.__orient == wx.VERTICAL:   orient = wx.HORIZONTAL
        elif self.__orient == wx.HORIZONTAL: orient = wx.VERTICAL

        self.InsertTool(ToolBarDivider(self, self.__height, orient), index)


    def AddTool(self, tool):
        """Adds the given tool to this ``FSLeyesToolBar``. """
        self.InsertTool(tool)


    def InsertTools(self, tools, index=None):
        """Inserts the given sequence of tools into this ``FSLeyesToolBar``,
        at the specified index.

        :arg tools: A sequence of tools to add.

        :arg index: Insert the tools before this index (default: end).
        """

        if index is None:
            index = self.GetToolCount()

        for i, tool in enumerate(tools, index):
            self.InsertTool(tool, i, postevent=False)

        wx.PostEvent(self, ToolBarEvent())


    def SetTools(self, tools, destroy=False):
        """Replaces all of the existing tools in this ``FSLeyesToolBar``
        with the given sequence of tools.

        :arg tools:   Sequence of new tools to add.

        :arg destroy: If ``True`` all of the old tools are destroyed.
        """

        self.ClearTools(destroy, postevent=False)

        for tool in tools:
            self.InsertTool(tool, postevent=False, redraw=False)

        self.__drawToolBar()

        wx.PostEvent(self, ToolBarEvent())


    def InsertTool(self, tool, index=None, postevent=True, redraw=True):
        """Inserts the given tool into this ``FSLeyesToolBar``, at the
        specified index.

        :arg tool:      The tool to insert.

        :arg index:     Index to insert the tool.

        :arg postevent: If ``True``, a :data:`ToolBarEvent` will be generated.
                        Pass ``False`` to suppress this event.

        :arg redraw:    If ``True``, the toolbar is redrawn. Pass ``False``
                        to suppress this behaviour.
        """

        if index is None:
            index = len(self.__tools)

        log.debug('{}: adding tool at index {}: {}'.format(
            type(self).__name__, index, type(tool).__name__))

        tool.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        # gtk3: something somewhere sometimes
        # clobbers the best size, so widgets
        # don't get shown. Only observed with
        # BitmapToggleButtons.
        size = tool.GetBestSize()
        tool.SetMinSize(size)
        tool.SetMaxSize(size)

        self.__tools.insert(index,     tool)
        self.__sizer.Insert(index + 1, tool, flag=wx.ALIGN_CENTRE)

        self.InvalidateBestSize()

        if redraw:
            self.__drawToolBar()

        if postevent:
            wx.PostEvent(self, ToolBarEvent())


    def DoGetBestSize(self):
        """Calculates and returns the best size for this toolbar, simply the
        minimum size that will fit all tools.

        This method is called by :mod:`wx` when this toolbar is laid out.
        """

        # Calculate the minimum/maximum size
        # for this toolbar, given the addition
        # of the new tool. If the orientation
        # of this toolbar (set in __init__) is
        # HORIZONTAL, the ttlSpace is used to
        # store total width, otherwise it is
        # used to store total height.
        ttlSpace  = 0
        minWidth  = 0
        minHeight = 0

        for tool in self.__tools:

            tw, th = tool.GetBestSize().Get()
            if tw > minWidth:  minWidth  = tw
            if th > minHeight: minHeight = th

            if self.__orient == wx.HORIZONTAL: ttlSpace += tw
            else:                              ttlSpace += th

        if self.__orient == wx.HORIZONTAL:
            leftWidth  = self.__leftButton .GetBestSize().GetWidth()
            rightWidth = self.__rightButton.GetBestSize().GetWidth()
            minWidth   = minWidth + leftWidth + rightWidth
        else:
            topHeight    = self.__leftButton .GetBestSize().GetHeight()
            bottomHeight = self.__rightButton.GetBestSize().GetHeight()
            minHeight    = minHeight + topHeight + bottomHeight

        if self.__orient == wx.HORIZONTAL: size = (ttlSpace, minHeight)
        else:                              size = (minWidth, ttlSpace)

        # The agw.AuiManager does not honour the best size when
        # toolbars are floated, but it does honour the minimum
        # size. So I'm just setting the minimum size to the best
        # size.
        log.debug('Setting toolbar size: {}'.format(size))

        self.SetMinSize(   size)
        self.SetMaxSize(   size)
        self.CacheBestSize(size)

        return size


    def ClearTools(
            self,
            destroy=False,
            startIdx=None,
            endIdx=None,
            postevent=True):
        """Removes all tools, or a range of tools, from this
        ``FSLeyesToolBar``.

        :arg destroy:   If ``True``, the removed tools are destroyed.

        :arg startIdx:  Start index of tools to remove. If not provided,
                        defaults to 0.

        :arg endIdx:    End index of tools to remove (exclusive). If not
                        provided, defaults to :meth:`GetToolCount()`.

        :arg postevent: If ``True``, a :data:`ToolBarEvent` will be
                        generated. Set to ``False`` to suppress the event.
        """

        if len(self.__tools) == 0:
            return

        if startIdx is None: startIdx = 0
        if endIdx   is None: endIdx   = len(self.__tools)

        for i in range(startIdx, endIdx):
            tool = self.__tools[i]

            self.__sizer.Detach(tool)

            if destroy:
                tool.Destroy()

        self.__tools[startIdx:endIdx] = []

        self.InvalidateBestSize()
        self.Layout()

        if postevent:
            wx.PostEvent(self, ToolBarEvent())


    def __onMouseWheel(self, ev):
        """Called when the mouse wheel is rotated on this ``FSLeyesToolBar``.

        Calls :meth:`__onLeftButton` or :meth:`__onRightButton`, depending
        on the rotation direction.
        """

        wheelDir = ev.GetWheelRotation()
        if   wheelDir < 0: self.__onRightButton()
        elif wheelDir > 0: self.__onLeftButton()


    def __onLeftButton(self, ev=None):
        """Called when the left toolbar button is pressed.

        If the toolbar is compressed, it is scrolled to the left.
        """

        self.__visibleOffset -= 1

        if self.__visibleOffset <= 0:
            self.__visibleOffset = 0

        log.debug('Left button pushed - setting start '
                  'tool index to {}'.format(self.__visibleOffset))

        self.__drawToolBar()


    def __onRightButton(self, ev=None):
        """Called when the right toolbar button is pressed.

        If the toolbar is compressed, it is scrolled to the right.
        """

        self.__visibleOffset += 1

        if self.__visibleOffset + self.__numVisible >= len(self.__tools):
            self.__visibleOffset = len(self.__tools) - self.__numVisible

        log.debug('Right button pushed - setting start '
                  'tool index to {}'.format(self.__visibleOffset))

        self.__drawToolBar()


    def __drawToolBar(self, *a):
        """Draws this ``FSLeyesToolBar``.

        If the toolbar is big enough, all tools are drawn. Otherwise, the
        method figures out out how many tools can be drawn, and which tools to
        draw, given the current size.
        """

        sizer  = self.__sizer
        tools  = self.__tools
        orient = self.__orient
        lbtn   = self.__leftButton
        rbtn   = self.__rightButton

        if orient == wx.HORIZONTAL:
            availSpace = self.GetSize().GetWidth()
            reqdSpace  = [tool.GetBestSize().GetWidth() for tool in tools]
            leftSpace  = lbtn .GetBestSize().GetWidth()
            rightSpace = rbtn .GetBestSize().GetWidth()

        else:
            availSpace = self.GetSize().GetHeight()
            reqdSpace  = [tool.GetBestSize().GetHeight() for tool in tools]
            leftSpace  = lbtn .GetBestSize().GetHeight()
            rightSpace = rbtn .GetBestSize().GetHeight()

        enoughSpace = availSpace >= sum(reqdSpace)

        sizer.Show(lbtn, not enoughSpace)
        sizer.Show(rbtn, not enoughSpace)

        # show all tools
        if enoughSpace:

            log.debug('{}: All tools fit ({} >= {})'.format(
                type(self).__name__, availSpace, sum(reqdSpace)))

            self.__visibleOffset = 0
            self.__numVisible    = len(tools)

            for tool in tools:
                sizer.Show(tool)

        # show <numVisible> tools, starting from <visibleOffset>
        # (see __onMouseWheel/__onLeftButton/__onRightButton)
        else:
            reqdSpace  = reqdSpace[self.__visibleOffset:]
            cumSpace   = np.cumsum(reqdSpace) + leftSpace + rightSpace
            biggerIdxs = [int(i) for i in np.where(cumSpace > availSpace)[0]]

            if len(biggerIdxs) == 0:
                lastIdx = len(tools)
            else:
                lastIdx = biggerIdxs[0] + self.__visibleOffset

            self.__numVisible = lastIdx - self.__visibleOffset

            log.debug('{}: {} tools fit ({} - {})'.format(
                type(self).__name__, self.__numVisible, self.__visibleOffset, lastIdx))

            lbtn.Enable(self.__visibleOffset > 0)
            rbtn.Enable(lastIdx < len(tools))

            for i in range(len(tools)):
                sizer.Show(tools[i], self.__visibleOffset <= i < lastIdx)

        self.Layout()


_ToolBarEvent, _EVT_TOOLBAR_EVENT = wxevent.NewEvent()


EVT_TOOLBAR_EVENT = _EVT_TOOLBAR_EVENT
"""Identifier for the :data:`ToolBarEvent` event. """


ToolBarEvent = _ToolBarEvent
"""Event emitted when one or more tools is/are added/removed to/from a
:class:`FSLeyesToolBar`.
"""


class ToolBarDivider(wx.Panel):
    """An empty ``wx.Panel`` intended to be used for dividing space in a
    :class:`FSLeyesToolBar`.
    """

    def __init__(self,
                 parent,
                 width=10,
                 height=32,
                 orient=wx.VERTICAL):

        wx.Panel.__init__(self, parent)

        if   orient == wx.VERTICAL:   size = (width,  height)
        elif orient == wx.HORIZONTAL: size = (height, width)

        self.SetMinSize(size)
        self.SetMaxSize(size)



#!/usr/bin/env python
#
# glrgbvolume_funcs.py - Functions used by the GLRGBVolume class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions used by the :class:`.GLRGBVolume` class
for rendering RGB(A) :class:`.Image` overlays in an OpenGL 2.1 environment.
"""


import fsleyes.colourmaps as fslcmaps
import fsleyes.gl.shaders as shaders
from . import                glvolume_funcs


def init(self):
    """Compiles and initialises the shader program. """
    self.shader = None
    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroys the shader program. """
    if self.shader is not None:
        self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Loads the vertex/fragment shader source code, and creates a
    :class:`.GLSLShader` program.
    """

    if self.shader is not None:
        self.shader.destroy()

    vertSrc = shaders.getVertexShader(  'glvolume')
    fragSrc = shaders.getFragmentShader('glrgbvolume')

    constants = {
        'textureIs2D' : self.imageTexture.ndim == 2
    }

    self.shader = shaders.GLSLShader(vertSrc, fragSrc, constants=constants)


def updateShaderState(self):
    """Updates all shader program variables. """

    if not self.ready():
        return

    display     = self.display
    opts        = self.opts
    image       = self.image
    shader      = self.shader
    imageShape  = image.shape[:3]
    nvals       = self.imageTexture.nvals
    texShape    = self.imageTexture.shape[:3]
    brightness  = display.brightness / 100
    contrast    = display.contrast   / 100
    rc, gc, bc  = self.channelColours()
    colourXform = fslcmaps.briconToScaleOffset(brightness, contrast, 1)
    hasAlpha    = not (nvals == 3 or opts.suppressA)

    if len(texShape) == 2:
        texShape = list(texShape) + [1]

    with shader.loaded():
        changed  = False
        changed |= shader.set('imageTexture',  0)
        changed |= shader.set('imageShape',    imageShape)
        changed |= shader.set('texShape',      texShape)
        changed |= shader.set('rcolour',       rc)
        changed |= shader.set('gcolour',       gc)
        changed |= shader.set('bcolour',       bc)
        changed |= shader.set('colourXform',   colourXform)
        changed |= shader.set('useSpline',     opts.interpolation == 'spline')
        changed |= shader.set('hasAlpha',      hasAlpha)

    return changed


def draw2D(self, *args, **kawrgs):
    """Draws a 2D slice at the given ``zpos``. Uses the
    :func:`.gl21.glvolume_funcs.draw2D` function.
    """
    with self.shader.loaded():
        glvolume_funcs.draw2D(self, *args, **kawrgs)


def drawAll(self, *args, **kawrgs):
    """Draws all specified slices. Uses the
    :func:`.gl21.glvolume_funcs.drawAll` function.
    """
    with self.shader.loaded():
        glvolume_funcs.drawAll(self, *args, **kawrgs)



#!/usr/bin/env python
#
# loadoverlay.py - Action which allows the user to load overlay files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadOverlayAction`, which allows the user
to load overlay files into the :class:`.OverlayList`.


This module also provides a collection of standalone functions which can be
called directly:

.. autosummary::
   :nosignatures:

   makeWildcard
   loadOverlays
   loadImage
   interactiveLoadOverlays


Finally, this module provides a singleton :class:`RecentPathManager` instance
called :attr:`recentPathManager`, which can be registered with to be notified
when new files have been loaded.
"""


import            logging
import            os
import os.path as op

import numpy   as np

import fsl.utils.idle               as idle
import fsl.utils.notifier           as notifier
import fsl.utils.settings           as fslsettings
import fsleyes_widgets.utils.status as status
import fsleyes.autodisplay          as autodisplay
import fsleyes.strings              as strings
import fsleyes.data                 as dutils
from . import                          base


log = logging.getLogger(__name__)


class LoadOverlayAction(base.Action):
    """The ``LoadOverlayAction`` allows the user to add files to the
    :class:`.OverlayList`.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``LoadOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__loadOverlay)


    def __loadOverlay(self):
        """Calls :func:`interactiveLoadOverlays`.

        If overlays were added, updates the
        :attr:`.DisplayContext.selectedOverlay` accordingly.

        If :attr:`.DisplayContext.autoDisplay` is ``True``, uses the
        :mod:`.autodisplay` module to configure the display properties
        of each new overlay.
        """

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)

            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        interactiveLoadOverlays(onLoad=onLoad,
                                inmem=self.displayCtx.loadInMemory)


def makeWildcard(allowedExts=None, descs=None):
    """Returns a wildcard string for use in a file dialog, to limit
    the the displayed file types to supported overlay file types.
    """

    import fsl.data.image          as fslimage
    import fsl.data.mghimage       as fslmgh
    import fsl.data.vtk            as fslvtk
    import fsl.data.gifti          as fslgifti
    import fsl.data.freesurfer     as fslfs
    import fsl.data.bitmap         as fslbmp
    import fsleyes.data.tractogram as tractogram

    # Hack - the wx wildcard logic doesn't support
    # files with multiple extensions (e.g. .nii.gz).
    # So I'm adding support for the '.gz' extension
    # here.
    fsfiles = [op.splitext(f)[1] for f in fslfs.CORE_GEOMETRY_FILES]
    if allowedExts is None:
        allowedExts  = (fslimage   .ALLOWED_EXTENSIONS +
                        fslvtk     .ALLOWED_EXTENSIONS +
                        fslmgh     .ALLOWED_EXTENSIONS +
                        fslgifti   .ALLOWED_EXTENSIONS +
                        tractogram .ALLOWED_EXTENSIONS +
                        fslbmp     .BITMAP_EXTENSIONS  +
                        fsfiles                        +
                        ['.gz'])
    if descs is None:
        descs        = (fslimage   .EXTENSION_DESCRIPTIONS     +
                        fslvtk     .EXTENSION_DESCRIPTIONS     +
                        fslmgh     .EXTENSION_DESCRIPTIONS     +
                        fslgifti   .EXTENSION_DESCRIPTIONS     +
                        tractogram .EXTENSION_DESCRIPTIONS     +
                        fslbmp     .BITMAP_DESCRIPTIONS        +
                        fslfs      .CORE_GEOMETRY_DESCRIPTIONS +
                        ['Compressed images'])

    exts  = ['*{}'.format(ext) for ext in allowedExts]
    exts  = [';'.join(exts)]        + exts
    descs = ['All supported files'] + descs

    wcParts = ['|'.join((desc, ext)) for (desc, ext) in zip(descs, exts)]

    return '|'.join(wcParts)


def loadOverlays(paths,
                 loadFunc='default',
                 errorFunc='default',
                 saveDir=True,
                 onLoad=None,
                 inmem=False,
                 blocking=False):
    """Loads all of the overlays specified in the sequence of files
    contained in ``paths``.

    .. note:: The overlays are loaded asynchronously via :func:`.idle.idle`.
              Use the ``onLoad`` argument if you wish to be notified when
              the overlays have been loaded.

    :arg loadFunc:  A function which is called just before each overlay
                    is loaded, and is passed the overlay path. The default
                    load function uses the :mod:`.status` module to display
                    the name of the overlay currently being loaded. Pass in
                    ``None`` to disable this default behaviour.

    :arg errorFunc: A function which is called if an error occurs while
                    loading an overlay, being passed the name of the
                    overlay, and either the :class:`Exception` which
                    occurred, or a string containing an error message.  The
                    default function pops up a :class:`wx.MessageBox` with
                    an error message. Pass in ``None`` to disable this
                    default behaviour.

    :arg saveDir:   If ``True`` (the default), the directory of the last
                    overlay in the list of ``paths`` is saved, and used
                    later on as the default load directory.

    :arg onLoad:    Optional function to call when all overlays have been
                    loaded. Must accept two parameters:

                     - a list of indices, one for each overlay, into the
                       ``paths`` parameter, indicating, for each overlay, the
                       path from which it was loaded.

                     - a list of the overlays that were loaded

    :arg inmem:     If ``True``, all :class:`.Image` overlays are
                    force-loaded into memory. Otherwise, large compressed
                    files may be kept on disk. Defaults to ``False``.

    :arg blocking:  Defaults to ``False``. If ``True``, overlays are loaded
                    immediately (and the ``onLoad`` function is called
                    directly. Otherwise, overlays and the ``onLoad`` are loaded
                    and called on the :func:`.idle.idle` loop.

    :returns:       If ``blocking is False`` (the default), returns ``None``.
                    Otherwise returns a list containing the loaded overlay
                    objects.
    """

    import fsl.data.image  as fslimage
    import fsl.data.mesh   as fslmesh
    import fsl.data.bitmap as fslbmp

    # The default load function updates
    # the dialog window created above
    def defaultLoadFunc(s):
        msg = strings.messages['loadOverlays.loading'].format(s)
        status.update(msg)

    # The default error function
    # shows an error dialog
    def defaultErrorFunc(s, e):
        status.reportError(
            strings.titles[  'loadOverlays.error'],
            strings.messages['loadOverlays.error'].format(s),
            e)

    # A function which loads a single overlay
    def loadPath(path, idx):

        loadFunc(path)

        dtype, path = dutils.guessType(path)

        if dtype is None:
            errorFunc(path, strings.messages['loadOverlays.unknownType'])
            return

        log.debug('Loading overlay {} (guessed data type: {})'.format(
            path, dtype.__name__))

        try:
            if   issubclass(dtype, fslimage.Image):
                loaded = loadImage(dtype, path, inmem=inmem)
            elif issubclass(dtype, fslmesh.Mesh):
                loaded = [dtype(path, fixWinding=True)]
            elif issubclass(dtype, fslbmp.Bitmap):
                loaded = [dtype(path).asImage()]
            else:
                loaded = [dtype(path)]

            overlays.extend(loaded)
            pathIdxs.extend([idx] * len(loaded))

        except Exception as e:
            errorFunc(path, e)

        # Record the path in the
        # recent files list
        recentPathManager.recordPath(path)

    # This function gets called after
    # all overlays have been loaded
    def realOnLoad(*a):

        if saveDir and len(paths) > 0:
            ovlDir = op.abspath(op.dirname(paths[-1]))
            fslsettings.write('loadSaveOverlayDir', ovlDir)

        if onLoad is not None:
            onLoad(pathIdxs, overlays)

    # If loadFunc or errorFunc are explicitly set to
    # None, use these no-op load/error functions
    if loadFunc  is None: loadFunc  = lambda s:    None
    if errorFunc is None: errorFunc = lambda s, e: None

    # Or if not provided, use the
    # default functions defined above
    if loadFunc  == 'default': loadFunc  = defaultLoadFunc
    if errorFunc == 'default': errorFunc = defaultErrorFunc

    pathIdxs = []
    overlays = []
    funcs    = []

    # Load the images
    for idx, path in enumerate(paths):
        funcs.append(lambda p=path, i=idx: loadPath(p, i))
    funcs.append(realOnLoad)

    for func in funcs:
        if blocking: func()
        else:        idle.idle(func)

    if blocking: return overlays
    else:        return None


def loadImage(dtype, path, inmem=False):
    """Called by the :func:`loadOverlays` function. Loads an overlay which
    is represented by an ``Image`` instance, or a sub-class of ``Image``.
    Depending upon the image size, the data may be loaded into memory or
    kept on disk, and the initial image data range may be calculated
    from the whole image, or from a sample.

    This function returns a sequence, most likely containing a single
    :class:`.Image` instance. But in some circumstances, more than one
    :class:`.Image` will be created and returned.

    :arg dtype: Overlay type (``Image``, or a sub-class of ``Image``).
    :arg path:  Path to the overlay file.
    :arg inmem: If ``True``, ``Image`` overlays are loaded into memory.
    :returns:   A sequence of :class:`.Image` instances that were loaded.
    """

    import fsl.data.image            as fslimage
    import fsleyes.data.imagewrapper as imagewrapper

    # We're going to load the file twice - first to
    # get its dimensions/data type, and then for real.
    #
    # TODO It is annoying that you have to create a 'dtype'
    #      instance twice, as e.g. the MelodicImage does a
    #      bunch of extra stuff (e.g. loading component
    #      time courses) that don't need to be done. Maybe
    #      the path passed to this function could be resolved
    #      (e.g. ./filtered_func.ica/ turned into
    #      ./filtered_func.ica/melodic_IC) so that you can
    #      just create a fsl.data.Image, or a nib.Nifti1Image.
    image  = dtype(path)
    shape  = image.shape
    nbytes = np.prod(shape) * image.niftiDataTypeSize / 8
    image  = None

    # If the file is a large 4D image, we will use an
    # ImageWrapper to manage acccess to the data. The
    # ImageWrapper will incrementally update the
    # known data range of an image as more data is
    # read from disk.
    useWrapper = fslsettings.read('fsleyes.overlay.usewrapper', 536870912)
    useWrapper = all((not inmem,
                      len(shape) > 3,
                      nbytes >= useWrapper))
    if useWrapper: wrapper = imagewrapper.ImageWrapper(threaded=True)
    else:          wrapper = None

    image = dtype(path, loadMeta=True, dataMgr=wrapper)

    # Force-load the full image data array
    # into memory. See the Image.data method.
    if inmem:
        image.data

    # If using an image wrapper, read a
    # sample of data to force the wrapper
    # to initialise its known data range
    if wrapper is not None:
        wrapper.setImage(image.nibImage)
        with wrapper.unthreaded():
            wrapper[..., 0]

    return [image]


def interactiveLoadOverlays(fromDir=None, dirdlg=False, **kwargs):
    """Convenience function for interactively loading one or more overlays.

    Pops up a file dialog prompting the user to select one or more overlays
    to load.

    :arg fromDir: Directory in which the file dialog should start.  If
                  ``None``, the most recently visited directory (via this
                  function) is used, or a directory from An already loaded
                  overlay, or the current working directory.

    :arg dirdlg:  Use a directory chooser instead of a file dialog.

    :arg kwargs:  Passed  through to the :func:`loadOverlays` function.

    :raise ImportError:  if :mod:`wx` is not present.
    :raise RuntimeError: if a :class:`wx.App` has not been created.
    """
    import wx

    app = wx.GetApp()

    if app is None:
        raise RuntimeError('A wx.App has not been created')

    saveFromDir = False
    if fromDir is None:

        saveFromDir = True
        fromDir     = fslsettings.read('loadSaveOverlayDir')

        if fromDir is None:
            fromDir = os.getcwd()

    if dirdlg:
        msg = strings.titles['interactiveLoadOverlays.dirDialog']
        dlg = wx.DirDialog(app.GetTopWindow(),
                           message=msg,
                           defaultPath=fromDir,
                           style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
    else:
        msg = strings.titles['interactiveLoadOverlays.fileDialog']
        dlg = wx.FileDialog(app.GetTopWindow(),
                            message=msg,
                            defaultDir=fromDir,
                            wildcard=makeWildcard(),
                            style=wx.FD_OPEN | wx.FD_MULTIPLE)

    if dlg.ShowModal() != wx.ID_OK:
        return

    if dirdlg: paths = [dlg.GetPath()]
    else:      paths =  dlg.GetPaths()

    dlg.Close()
    dlg.Destroy()

    loadOverlays(paths, saveDir=saveFromDir, **kwargs)


class RecentPathManager(notifier.Notifier):
    """The ``RecentPathManager`` is a simple class which provides
    access to a list of recently loaded files, and can notify
    registered listeners when that list changes. See the
    :attr:`recentPathManager` singleton instance.
    """


    def recordPath(self, path):
        """Adds the given ``path`` to the recent files list. """

        recent = self.listRecentPaths()

        if path in recent:
            return

        recent.append(path)

        if len(recent) > 10:
            recent = recent[-10:]

        recent = op.pathsep.join(recent)

        fslsettings.write('fsleyes.recentFiles', recent)

        self.notify()


    def listRecentPaths(self):
        """Returns a list of recently loaded files. """

        recent = fslsettings.read('fsleyes.recentFiles', None)

        if recent is None: recent = []
        else:              recent = recent.split(op.pathsep)

        return [f for f in recent if op.exists(f)]


recentPathManager = RecentPathManager()
"""A :class:`RecentPathManager` instance which gets updated by the
:func:`loadOverlays` function whenever a new path is loaded. Register
as a listener on this instance if you want to be notified of changes
to the recent paths list.
"""



#!/usr/bin/env python
#
# tooltips.py - Tooltips for FSLeyes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains tooltips used throughout *FSLeyes*.

Tooltips are stored in :class:`.TypeDict` dictionariesa, broadly organised
into the following categories:

 ================== ================================================
 :data:`properties` Tooltips for ``props.HasProperties`` properties.
 :data:`actions`    Tooltips for :class:`.ActionProvider` actions.
 :data:`misc`       Tooltips for everything else.
 ================== ================================================

The :func:`initTooltips` function initialises some parameters controlling
tooltip display. It is called by the :class:`.FSLeyesFrame` upon creation.
"""


from fsleyes_widgets.utils.typedict import TypeDict


def initTooltips():
    """Sets some parameters controlling tooltip display. """
    import wx
    wx.ToolTip.Enable(     True)
    wx.ToolTip.SetDelay(   1500)
    wx.ToolTip.SetMaxWidth(300)
    wx.ToolTip.SetReshow(  3000)
    wx.ToolTip.SetAutoPop( 5000)


properties = TypeDict({

    # DisplayContext

    'DisplayContext.displaySpace'     : 'The space in which overlays are '
                                        'displayed.',
    'DisplayContext.radioOrientation' : 'If checked, images oriented to the '
                                        'MNI152 standard will be displayed in '
                                        'radiological orientation (i.e. with '
                                        'subject left to the right of the '
                                        'display, and subject right to the '
                                        'left). Otherwise they will be '
                                        'displayed in neurological '
                                        'orientation (i.e. with subject left '
                                        'to the left of the display).',

    # Overlay Display

    'Display.name'        : 'The name of this overlay.',
    'Display.overlayType' : 'The overlay type - how this overlay should be '
                            'displayed.',
    'Display.enabled'     : 'Show/hide this overlay.',
    'Display.alpha'       : 'The opacity of this overlay.',
    'Display.brightness'  : 'The brightness of this overlay. For volume '
                            'overlays, brightness is applied as a linear '
                            'offset to the display range.',
    'Display.contrast'    : 'The contrast of this overlay. For volume '
                            'overlay, contrast is applied as a linear '
                            'scaling factor to the display range.',

    # Overlay DisplayOpts

    'NiftiOpts.volume'     : 'The volume number (for 4D images).',
    'NiftiOpts.transform'  : 'The affine transformation matrix to apply '
                             'to this image. You can choose to display '
                             'the image without any transformation (as if '
                             'the image voxels are 1mm isotropic); or you '
                             'can choose to scale the voxels by the pixdim '
                             'values in the NIFTI header; or you can choose '
                             'to apply the affine transformation as defined '
                             'in the NIFTI header.',


    'ColourMapOpts.displayRange' :
    'Data display range - the low value corresponds to the low colour, and '
    'the high value to the high colour, in the selected colour map.',

    'ColourMapOpts.clippingRange' :
    'Data clipping range - voxels with values outside of this range will not '
    'be displayed.',

    'ColourMapOpts.modulateRange' :
    'Modulation range - controls the data range used to modulate transparency',

    'ColourMapOpts.invertClipping' :
    'Invert the clipping range, so that voxels inside the range are not '
    'displayed, and voxels outside of the range are displayed. '
    'This option is useful for displaying statistic images.',

    'ColourMapOpts.cmap'            : 'The colour map to use.',
    'ColourMapOpts.custom_cmap'     : 'The colour map to use.',

    'ColourMapOpts.negativeCmap' :
    'The colour map to use for negative values.',

    'ColourMapOpts.useNegativeCmap' :
    'Enable the negative colour map - this allows positive and negative '
    'values to be coloured independently.',

    'ColourMapOpts.cmapResolution'  :
    'Colour map resolution - the number of colours to use in the colour maps.',

    'ColourMapOpts.interpolateCmaps' :
    'Interpolate between discrete colours in the colour map. If not selected, '
    'nearest-neighbour interpolation is used. ',

    'ColourMapOpts.invert' :
    'Invert the display range, so that the low value corresponds to the high '
    'colour, and vice versa.',

    'ColourMapOpts.modulateAlpha' :
    'Modulate alpha (opacity) by the intensity at each region. Regions with a '
    'value near to the low display range will have opacity near 0, and '
    'regions with a value near to the high display range will have opacity '
    'near 1.',

    'ColourMapOpts.gamma' :
    'Apply an exponential factor to to give more weighting to lower or higher '
    'values. ',

    'ColourMapOpts.logScale' :
    'Map intensities to colours logarithmically (using the natural logarithm) '
    'rather than linearly. ',

    'VolumeOpts.interpolation' :
    'Interpolate the image data on the display. You can choose no  '
    'interpolation (equivalent to nearest neighbour interpolation), linear '
    'interpolation, or third-order spline (cubic) interpolation.',

    'VolumeOpts.enableOverrideDataRange' :
    'Override the actual data range of an image with a user-specified '
    'one. This is useful for images which have a very large data '
    'range that is driven by outliers.',

    'VolumeOpts.overrideDataRange' :
    'Override the actual data range of an image with a user-specified '
    'one. This is useful for images which have a very large data '
    'range that is driven by outliers.',

    'VolumeOpts.custom_overrideDataRange' :
    'Override the actual data range of an image with a user-specified '
    'one. This is useful for images which have a very large data '
    'range that is driven by outliers.',

    'VolumeOpts.clipImage' :
    'Clip this image by the values contained in another image. When active, '
    'The clipping range is set according to the values in the clip image '
    'instead of in this image.',

    'VolumeOpts.modulateImage' :
    'Modulate alpha/opacity by the values contained in another image, instead '
    'of modulating by the values in this image.',

    'Volume3DOpts.numSteps' :
    'The maximum number of times that the image is sampled for each pixel.',

    'Volume3DOpts.blendFactor' :
    'This setting controls the sampling depth - a higher value will result '
    'in the final volume colour being blended from samples which are deeper '
    'in the volume.',

    'Volume3DOpts.blendByIntensity' :
    'When active, the amount by which each sample is blended into the final '
    'colour will be weighted by the voxel intensity. When inactive, samples '
    'will be blended solely according to the blending setting.',

    'Volume3DOpts.smoothing' :
    'This setting controls the amount of smoothing applied to the volume. '
    'When the smoothing is greater than zero, a gaussian smoothing filter is '
    'applied, with the filter radius (in pixels) controlled by this setting.',

    'Volume3DOpts.resolution' :
    'This setting controls the quality (resolution) at which the volume is '
    'shown on screen. Higher values look better, but lower values will be '
    'drawn more quickly.',

    'Volume3DOpts.numClipPlanes' :
    'Number of active clipping planes. Areas of the image which are in the '
    'intersection, union, or complement of all clipping planes will not be '
    'shown. The clip mode controls how clipping is applied',

    'Volume3DOpts.clipMode' :
    'How the clipping planes are applied. You can choose to clip (hide) the '
    'intersection, union, or complement of all active clipping planes',

    'Volume3DOpts.showClipPlanes' :
    'When enabled, each active clipping plane is shown.',

    'Volume3DOpts.clipPosition' :
    'Clip plane position, as a percentage of the image length.',

    'Volume3DOpts.clipInclination' :
    'Amount by which to rotate the clip plane about the X axis.',

    'Volume3DOpts.clipAzimuth' :
    'Amount by which to rotate the clip plane about the Z axis.',

    'ComplexOpts.component' :
    'Choose to display either the real or imaginary components, or the '
    'magnitude or phase.',

    'MaskOpts.colour' :
    'The colour of this mask image.',
    'MaskOpts.invert' :
    'Invert the mask threshold range, so that values outside of the range '
    'are shown, and values within the range are hidden.',
    'MaskOpts.threshold' :
    'The mask threshold range - values outside of this range will not be '
    'displayed.',
    'MaskOpts.outline' :
    'When selected, an outline of the mask is shown, instead of the filled '
    'mask being shown.',
    'MaskOpts.outlineWidth'  :
    'When the mask outline is shown, this setting controls the outline width '
    'in pixels.',
    'MaskOpts.interpolation' :
    'Interpolate the mask data on the display. You can choose no  '
    'interpolation (equivalent to nearest neighbour interpolation), linear '
    'interpolation, or third-order spline (cubic) interpolation.',

    'LabelOpts.lut'          : 'The lookup table to use for this label image.',
    'LabelOpts.outline'      : 'Show the outline of each labelled region '
                               'only. If unchecked, labelled regions are '
                               'filled.',
    'LabelOpts.outlineWidth' : 'If showing label outlines, this setting '
                               'controls the outline width (as a proportion '
                               'of the image voxel size). If showing filled '
                               'regions, this setting controls the size of a '
                               'transparent border around each region. In '
                               'this situation, setting the width to 0 will '
                               'prevent the border from being shown.',
    'LabelOpts.showNames'    : 'Annotate the image display with the names of '
                               'each labelled region.',

    'VectorOpts.xColour'          : 'The colour corresponding to the X '
                                    'component of the vector - the brightness '
                                    'of the colour corresponds to the '
                                    'magnitude of the X component. This '
                                    'option has no effect if a colour image '
                                    'is selected.',
    'VectorOpts.yColour'          : 'The colour corresponding to the Y '
                                    'component of the vector - the brightness '
                                    'of the colour corresponds to the '
                                    'magnitude of the Y component. This '
                                    'option has no effect if a colour image '
                                    'is selected.',
    'VectorOpts.zColour'          : 'The colour corresponding to the Z '
                                    'component of the vector - the brightness '
                                    'of the colour corresponds to the '
                                    'magnitude of the Z component. This '
                                    'option has no effect if a colour image '
                                    'is selected.',
    'VectorOpts.suppressX'        : 'Ignore the X vector component when '
                                    'colouring voxels. This option has no '
                                    'effect if a colour image is selected.',
    'VectorOpts.suppressY'        : 'Ignore the Y vector component when '
                                    'colouring voxels. This option has no '
                                    'effect if a colour image is selected.',
    'VectorOpts.suppressZ'        : 'Ignore the Z vector component when '
                                    'colouring voxels. This option has no '
                                    'effect if a colour image is selected.',
    'VectorOpts.suppressMode'     : 'When a vector direction is suppressed,'
                                    'it\'s contribution to the resulting '
                                    'will be replaced according to this '
                                    'setting.',
    'VectorOpts.modulateImage'    : 'Modulate the vector colour brightness by '
                                    'another image. The image selected here '
                                    'is normalised to lie in the range (0, '
                                    '1), and the brightness of each vector '
                                    'colour is scaled by the corresponding '
                                    'modulation value before it is coloured. '
                                    'The modulation image must have the same '
                                    'voxel dimensions as the vector image.',
    'VectorOpts.clipImage'        : 'Clip vector voxels according to the '
                                    'values in another image. Vector voxels '
                                    'which correspond to values in the '
                                    'clipping image that have a value less '
                                    'than the current clipping threshold are '
                                    'not shown. The clipping image must have '
                                    'the same voxel dimensions as the vector '
                                    'image. ',
    'VectorOpts.colourImage'      : 'Colour the vectors according to the '
                                    'values in another image, and by the '
                                    'selected colour map. The colour image '
                                    'must have the same voxel dimensions as '
                                    'the vector image. ',
    'VectorOpts.clippingRange'    : 'Vector values which have a corresponding '
                                    'clipping image value that is outside of '
                                    'this range are not displayed. ',
    'VectorOpts.modulateRange'    : 'The data range that is used when '
                                    'modulating vector brightness by a '
                                    'modulation image.',
    'VectorOpts.modulateMode'     : 'Either the brightness, or the '
                                    'transparency, of vector voxels can be '
                                    'modulated by the modulation image.',
    'VectorOpts.cmap'             : 'Colour map to use for colouring vector '
                                    'voxels, if a colour image is selected.',
    'VectorOpts.orientFlip'       : 'If checked, direction orientations '
                                    'within each voxel are flipped about '
                                    'the x axis.',
    'LineVectorOpts.lineWidth'    : 'The width of each vector line, in '
                                    'display pixels.',
    'LineVectorOpts.directed'     : 'If unchecked, the vector data is assumed '
                                    'to be undirected - the vector line at '
                                    'each voxel is scaled to have a length of '
                                    '1mm, and is centered within the voxel so '
                                    'that it passes through the voxel centre.'
                                    'If this option is checked, the vector '
                                    'data is assumed to be directed - each '
                                    'vector line begins at the voxel centre, '
                                    'and is scaled to have length 0.5mm.',
    'LineVectorOpts.unitLength'   : 'If checked, the vector lines are scaled '
                                    'so that they have a length of 1mm (times'
                                    'the length scaling factor). '
                                    'Otherwise the vector lengths are '
                                    'unmodified.',
    'LineVectorOpts.lengthScale'  : 'Scale the vector line length by this '
                                    'scaling factor (expressed as a '
                                    'percentage).',
    'RGBVectorOpts.interpolation' : 'Interpolate the vector data for display '
                                    'purposes. You can choose none '
                                    '(equivalent to nearest-neighbour), '
                                    'linear, or spline interpolation.',
    'RGBVectorOpts.unitLength'    : 'If checked, the data is scaled so that '
                                    'each vector has length 1.',


    'MeshOpts.colour' :
    'The colour of the mesh, when not colouring by vertex data.',
    'MeshOpts.outline' :
    'If checked, only the outline of the mesh is displayed. Otherwise the '
    'mesh cross-section is shown. ',
    'MeshOpts.outlineWidth' :
    'If the mesh outline is being displayed, this setting controls the '
    'outline width.',
    'MeshOpts.showName' :
    'Annotate the display wiuh the model name.',
    'MeshOpts.refImage' :
    'If this model was derived from a volumetric image, you can choose that '
    'image as a reference. The displayed model will then be transformed '
    'according to the transformation/orientation settings of the reference.',
    'MeshOpts.coordSpace'   :
    'If a reference image is selected, this setting defines the space, '
    'relative to the reference image, in which the model coordinates are '
    'defined.',
    'MeshOpts.vertexSet' :
    'Choose a file which contains a surface definition (vertices) for this '
    'mesh',
    'MeshOpts.custom_vertexSet' :
    'Choose a file which contains a surface definition (vertices) for this '
    'mesh',
    'MeshOpts.vertexData' :
    'Choose a file which contains data for each vertex - you can colour the '
    'mesh outline according to the values in the file. This only applies '
    'when the mesh outline, and not its cross-section is displayed.',
    'MeshOpts.modulateData' :
    'Choose a vertex data file to modulate transparency by. If left unset, '
    'transparency is modulated by the currently selected vertex data.',
    'MeshOpts.custom_vertexData'   :
    'Choose a file which contains data for each vertex - you can colour the '
    'mesh outline according to the values in the file. This only applies '
    'when the mesh outline, and not its cross-section is displayed.',
    'MeshOpts.vertexDataIndex' :
    'If you have loaded vertex data with multiple data points for each '
    'vertex, this control allows you to control the data point that is '
    'displayed.',
    'MeshOpts.useLut' :
    'When selected, a lookup table, instead of the colour maps, is used to '
    'colour the mesh according to its vertex data. This is useful for '
    'discrete label data',
    'MeshOpts.lut' :
    'The lookup table to use when colouring the mesh with a lookup table '
    'instead of with the colour maps.',
    'MeshOpts.custom_lut' :
    'The lookup table to use when colouring the mesh with a lookup table '
    'instead of with the colour maps.',
    'MeshOpts.discardClipped' :
    'When the mesh is coloured according to some data, this setting allows '
    'you to choose between hiding the areas with data outside of the clipping '
    'range, or colouring those areas with a constant colour.',
    'MeshOpts.wireframe' :
    'When selected, the mesh is shown as a wireframe, rather than being '
    'filled.',
    'MeshOpts.lighting' :
    'When selected, a lighting effect is applied to the mesh.',
    'MeshOpts.interpolation' :
    'Interpolation of vertex data values - when linear interpolation is '
    'selected, each triangle is coloured according to the vertex data values '
    'at each vertex. When nearest neighbour interpolation is selected, each '
    'triangle in the mesh is coloured according to its first triangle.',

    'TensorOpts.lighting'         : 'If enabled, a simple lighting model is '
                                    'used to highlight the tensor '
                                    'orientations.',
    'TensorOpts.tensorResolution' : 'This setting controls the number of '
                                    'vertices used to render each tensor. '
                                    'A higher value will result in better '
                                    'looking tensors, but may reduce '
                                    'performance.' ,
    'TensorOpts.tensorScale'      : 'By default, the tensor radii are scaled '
                                    'the largest eigenvalue of the tensor '
                                    'matrix, so that the largest tensor is '
                                    'drawn to fit within a voxel. This '
                                    'setting allows the tensor scale to be '
                                    'adjusted.',

    'SHOpts.lighting'         : 'If enabled, a simple lighting model is used '
                                'to highlight the FODs',
    'SHOpts.normalise'        : 'If enabled, the size of each FOD is '
                                'normalised to fit within a voxel.',
    'SHOpts.size'             : 'This setting allows the FOD size to be '
                                'scaled up or down',
    'SHOpts.radiusThreshold'  : 'This setting allows FODs with small radius '
                                'to be hidden.',
    'SHOpts.shResolution'     : 'This setting controls the display resolution '
                                '(number of vertices) used to draw each FOD.',
    'SHOpts.shOrder'          : 'This setting controls the maximum spherical '
                                'harmonic function order with which to '
                                'display FODs.',
    'SHOpts.colourMode'       : 'FODs can be coloured according to their '
                                'radius/size, or according to their '
                                'orientation/direction. This setting is '
                                'disabled when you choose to colour the FODs '
                                'by another image (e.g. a FA map).',

    'TractogramOpts.colourMode' :
    'Tractograms can be coloured according to their orientation, according to '
    'data taken fraom a NIfTI image, or according to per-vertex/streamline '
    'values loaded from a file, or contained within the tractogram file.',
    'TractogramOpts.clipMode' :
    'Tractograms can be clipped according to data taken from a NIfTI image, '
    'or according to  per-vertex/streamline values loaded from a file, or '
    'contained within the tractogram file.',
    'TractogramOpts.lineWidth' :
    'Controls the thickness of streamline lines/tubes (when viewing in 3D), '
    'or the diameter of streamline vertex points (when viewing in 2D).',
    'TractogramOpts.resolution' :
    'Controls the quality when drawing streamline tubes (in 3D) or vertex '
    'points (in 2D).',
    'TractogramOpts.subsample' :
    'Draw a randomly selected subset of streamlines. This setting is useful '
    'when viewing very large tractograms.',

    # MIPOpts
    'MIPOpts.window'  :
    'Length of the window, as a proportion of the image length, along which '
    'the MIP is calculated. The window is centred at the current display '
    'location.',
    'MIPOpts.minimum' :
    'Use the minimum intensity, rather than the maximum intensity, in the '
    'projection.',
    'MIPOpts.absolute' :
    'Use the absolute intensity, rather than the maximum intensity, in the '
    'projection. This overrides the minimum intensity setting.',
    'MIPOpts.interpolation' :
    'Interpolate the MIP data on the display. You can choose no  '
    'interpolation (equivalent to nearest neighbour interpolation), linear '
    'interpolation, or third-order spline (cubic) interpolation.',

    # SceneOpts
    'SceneOpts.showCursor'         : 'Show/hide the cursor which highlights '
                                     'the current location.',
    'SceneOpts.cursorWidth'        : 'Width/thickness of the location cursor.',
    'SceneOpts.cursorGap'          : 'Show a gap at the cursor centre.',
    'SceneOpts.cursorColour'       : 'Colour of the location cursor.',
    'SceneOpts.bgColour'           : 'Canvas background colour.',
    'SceneOpts.fgColour' :
    'Foreground colour, used for labels and the colour bar. Note that the '
    'foreground colour will be automatically adjusted whenever you change '
    'the canvas background colour.',

    'SceneOpts.showColourBar'      : 'If the currently selected overlay is a '
                                     'volumetric image, show a colour bar '
                                     'depicting the colour/data display '
                                     'range.',
    'SceneOpts.colourBarLocation'  : 'Where to display the colour bar.',
    'SceneOpts.colourBarLabelSide' : 'What side of the colour bar to draw the '
                                     'colour bar labels.',
    'SceneOpts.colourBarSize'      : 'Size of the major axis of the colour '
                                     'bar, as a proportion of the available '
                                     'space.',
    'SceneOpts.highDpi' :
    'If you are using a high-DPI (e.g. retina) display, FSLeyes will attempt '
    'to display the scene at the high-DPI resolution. This will have an '
    'effect on performance.',

    'SceneOpts.labelSize'   : 'Scale the label font size.',


    'OrthoOpts.showXCanvas' : 'Show / hide the X canvas '
                              '(sagittal in MNI space).',
    'OrthoOpts.showYCanvas' : 'Show / hide the Y canvas '
                              '(coronal in MNI space).',
    'OrthoOpts.showZCanvas' : 'Show / hide the Z canvas '
                              '(axial in MNI space).',
    'OrthoOpts.showCursor'  : 'Show/hide the location cross-hairs.',
    'OrthoOpts.showLabels'  : 'If the currently selected overlay is a NIFTI '
                              'image, show / hide anatomical orientation '
                              'labels.',
    'OrthoOpts.showLocation' :
    'Show the cursor coordinates on one of the three canvases.',
    'OrthoOpts.layout'      : 'How to lay out each of the three canvases.',
    'OrthoOpts.zoom'        : 'Zoom level for all three canvases.',

    'OrthoToolBar.showCursorAndLabels' :
    'Show/hide the location cursor and anatomical labels',

    'Scene3DToolBar.showCursorAndLegend' :
    'Show/hide the location cursor and anatomical legend',

    'LightBoxOpts.zoom'           : 'Zoom level - this controls how many '
                                    'slices to display.',
    'LightBoxOpts.sliceSpacing'   : 'The spacing between adjacent slices. '
                                    'The units and range of this setting '
                                    'depend upon the currently selected '
                                    'overlay.',
    'LightBoxOpts.zrange'         : 'The start/end points of the displayed '
                                    'range of slices. The units and range '
                                    'of this setting depend upon the '
                                    'currently selected overlay.',
    'LightBoxOpts.zax'            : 'Slices along this axis will be '
                                    'displayed.',
    'LightBoxOpts.showGridLines'  : 'If checked, lines will be shown between '
                                    'each slice.',
    'LightBoxOpts.highlightSlice' : 'If checked, a box will be drawn around '
                                    'the currently selected slice.',

    'Scene3DOpts.showLegend' :
    'When selected, a legend will be displayed in the bottom right, showing '
    'orientation information.',
    'Scene3DOpts.light'      :
    'Enable a simple lighting effect.',
    'Scene3DOpts.showLight' :
    'Show the position of the light source.',
    'Scene3DOpts.lightPos' :
    'Position of the light in the display coordinate system.',
    'Scene3DOpts.lightDistance' :
    'Distance of the light from the display centre.',

    'Scene3DOpts.zoom' :
    'Zoom level - distance from the camera to the model space.',

    # ViewPanels

    'CanvasPanel.syncLocation'       : 'If checked, the location shown on '
                                       'this panel will be linked to the '
                                       'location shown on other panels (as '
                                       'long as they also have this setting '
                                       'enabled).',
    'CanvasPanel.syncOverlayOrder'   : 'If checked, the order in which '
                                       'overlays are displayed on this '
                                       'panel will be linked to order shown '
                                       'on other panels (as long as they '
                                       'also have this setting enabled). ',
    'CanvasPanel.syncOverlayDisplay' : 'If checked, the display properties '
                                       'of all overlays shown in this panel '
                                       'are linked to the display properties '
                                       'on other panels (as long as they '
                                       'also have this setting enabled). ',
    'CanvasPanel.syncOverlayVolume'  : 'If checked,  properties which control '
                                       'the displayed volume/timepoint '
                                       'of all overlays shown in this panel '
                                       'are linked to the volume properties '
                                       'on other panels (as long as they '
                                       'also have this setting enabled). ',

    'CanvasPanel.movieMode' :
    'If checked, the volume will automatically change at a rate determined '
    'by the movie rate.  You can also loop through the or X, Y, or Z voxel '
    'coordinates by changing the movie axis  inthe view settings panel.  If '
    'you are looping through volumes and you want several overlays to be '
    'animated, group them using the overlay list.',

    'CanvasPanel.movieRate' :
    'The rate at which volumes are changed when movie mode is enabled. Low = '
    'fast, and high = slow.',

    'CanvasPanel.movieAxis' :
    'The axis to loop through. You can choose to run a movie through any '
    'axis of an image (X, Y, Z, or volume/time).',

    'CanvasPanel.movieSyncRefresh' :
    'In movie mode, refresh all orthographic canvases in a synchronous '
    'manner. This is not possible under certain platforms/environments, '
    'where the canvas updates cannot be synchronised.',

    'PlotCanvas.legend'     : 'Show / hide a legend for series which have '
                             'been added to the plot.',
    'PlotCanvas.xAutoScale' : 'If checked, the plot X axis limits are '
                             'automatically adjusted whenever the plot '
                             'contents change.',
    'PlotCanvas.yAutoScale' : 'If checked, the plot Y axis limits are '
                             'automatically adjusted whenever the plot '
                             'contents change.',
    'PlotCanvas.xLogScale'  : 'If checked, a log (base 10) scale is used for '
                             'the x axis.',
    'PlotCanvas.yLogScale'  : 'If checked, a log (base 10) scale is used for '
                             'the y axis.',
    'PlotCanvas.invertX'    : 'Invert the plot along the X axis.',
    'PlotCanvas.invertY'    : 'Invert the plot along the Y axis.',
    'PlotCanvas.ticks'      : 'Show / hide axis ticks and tick labels.',
    'PlotCanvas.grid'       : 'Show hide plot grid.' ,
    'PlotCanvas.gridColour' : 'Set the plot grid colour.' ,
    'PlotCanvas.bgColour'   : 'Set the plot background colour.' ,
    'PlotCanvas.smooth'     : 'Smooth displayed data series (with cubic spline '
                             'interpolation).',
    'PlotCanvas.xlabel'     : 'Set the x axis label.',
    'PlotCanvas.ylabel'     : 'Set the y axis label.',
    'PlotCanvas.limits'     : 'Manually set the x/y axis limits.',

    'TimeSeriesPanel.usePixdim'        : 'If checked, the x axis data is '
                                         'scaled by the time dimension pixdim '
                                         'value specified in the NIFTI '
                                         'header.',
    'TimeSeriesPanel.plotMelodicICs'   : 'If checked, the component time '
                                         'courses are plotted for Melodic '
                                         'images. If not checked, Melodic '
                                         'images are treated as regular 4D '
                                         'images.',
    'TimeSeriesPanel.plotMode'         : 'Plotting mode. You can choose to: '
                                         '\n  - Display the data as-is.'
                                         '\n  - Remove the temporal mean from '
                                         'the data before plotting.'
                                         '\n  - Scale the data to the range '
                                         '[-1, 1].'
                                         '\n  - Scale the data to percent '
                                         'signal-changed, relative to the '
                                         'temporal mean.',
    'TimeSeriesPanel.currentColour'    : 'Colour of the current time series.',
    'TimeSeriesPanel.currentAlpha'     : 'Opacity of the current time series.',
    'TimeSeriesPanel.currentLineWidth' : 'Line width of the current time '
                                         'series.',
    'TimeSeriesPanel.currentLineStyle' : 'Line style of the current time '
                                         'series.',

    'HistogramPanel.histType' :
    'Show histogram data as raw counts, or as probabilities.',
    'HistogramPanel.plotType'    :
    'Use histogram bin edges or bin centres for the histogram plot.',

    'PowerSpectrumPanel.plotFrequencies'  : 'If checked, the x values '
                                            'are transformed into frequency '
                                            'values.',
    'PowerSpectrumPanel.plotMelodicICs'   : 'If checked, the component power '
                                            'spectra are plotted for Melodic '
                                            'images. If not checked, Melodic '
                                            'images are treated as regular 4D '
                                            'images.',

    # DataSeries

    'DataSeries.enabled'   : 'Show/hide the line.',
    'DataSeries.colour'    : 'Line colour.',
    'DataSeries.alpha'     : 'Line opacity.',
    'DataSeries.label'     : 'Line label (shown in the legend).',
    'DataSeries.lineWidth' : 'Line width.',
    'DataSeries.lineStyle' : 'Line style.',

    'ComplexTimeSeries.plotReal' :
    'Plot the real component of a complex image.',
    'ComplexTimeSeries.plotImaginary' :
    'Plot the imaginary component of a complex image.',
    'ComplexTimeSeries.plotMagnitude' :
    'Plot the magnitude of a complex image.',
    'ComplexTimeSeries.plotPhase' :
    'Plot the phase of a complex image.',

    'ComplexPowerSpectrumSeries.plotReal' :
    'Plot the real component of the power spectrum of a complex image.',
    'ComplexPowerSpectrumSeries.plotImaginary' :
    'Plot the imaginary component of the power spectrum of a complex image.',
    'ComplexPowerSpectrumSeries.plotMagnitude' :
    'Plot the magnitude of the power spectrum of a complex image.',
    'ComplexPowerSpectrumSeries.plotPhase' :
    'Plot the phase of the power spectrum of a complex image.',
    'ComplexPowerSpectrumSeries.zeroOrderPhaseCorrection' :
    'Zero order phase correction',
    'ComplexPowerSpectrumSeries.firstOrderPhaseCorrection' :
    'First order phase correction',

    'ComplexHistogramSeries.plotReal' :
    'Plot the histogram of the real component of a complex image.',
    'ComplexHistogramSeries.plotImaginary' :
    'Plot the histogram of the imaginary component of a complex image.',
    'ComplexHistogramSeries.plotMagnitude' :
    'Plot the histogram of the magnitude of a complex image.',
    'ComplexHistogramSeries.plotPhase' :
    'Plot the histogram of the phase of a complex image.',

    'FEATTimeSeries.plotData'         : 'Plot the input data.',
    'FEATTimeSeries.plotFullModelFit' : 'Plot the full model fit.',
    'FEATTimeSeries.plotResiduals'    : 'Plot the residuals of the full '
                                        'model fit.',
    'FEATTimeSeries.plotEVs'          : 'Plot the EV (explanatory variable) '
                                        'time courses.',
    'FEATTimeSeries.plotPEFits'       : 'Plot the model fit to each PE '
                                        '(parameter estimate).',
    'FEATTimeSeries.plotCOPEFits'     : 'Plot the model fit to each COPE '
                                        '(Contrast of Parameter Estimates).',
    'FEATTimeSeries.plotPartial'      : 'Plot the raw data, after regression '
                                        'against the selected PE/COPE.',

    'HistogramSeries.autoBin'         : 'If checked, automatically calculate '
                                        'the number of bins to use in the '
                                        'histogram calculation.',
    'HistogramSeries.nbins'           : 'Number of bins to use in the '
                                        'histogram calculation (not '
                                        'applicable  if auto-binning is '
                                        'enabled).',
    'HistogramSeries.ignoreZeros'     : 'Ignore zeros in the histogram '
                                        'calculation.',
    'HistogramSeries.showOverlay'     : 'Show a 3D mask overlay highlighting '
                                        'voxels which have been included in '
                                        'the histogram.',
    'HistogramSeries.includeOutliers' : 'Include values which are outside of '
                                        'the data range - they are added to '
                                        'the first and last bins.',
    'HistogramSeries.volume'          : 'Current volume to calculate the '
                                        'histogram for (4D images only).',
    'HistogramSeries.dataRange'       : 'Data range to include in the '
                                        'histogram.',


    'PowerSpectrumSeries.varNorm' :
    'If checked, the fourier-transformed data is normalised to the range '
    '[-1, 1]. Complex valued data are normalised with respect to the '
    'absolute value. ',

    # Profiles
    'OrthoPanel.profile' :
    'Switch between view mode and edit mode',

    'OrthoEditProfile.selectionCursorColour'  :
    'Colour to use for the selection cursor.',

    'OrthoEditProfile.selectionOverlayColour' :
    'Colour to use to highlight selected regions.',

    'OrthoEditProfile.locationFollowsMouse' :
    'Change the cursor location when you click and drag to draw, erase, or '
    'select voxels. If you are using a slower computer you may wish to '
    'disable this option.',

    'OrthoEditProfile.showSelection' :
    'Show/hide the current selection, when in select mode.',

    'OrthoEditProfile.drawMode' :
    'Toggle between "draw" mode and "select" mode. In draw mode, you can '
    'simply \'draw\' on an image - when you release the mouse, the voxel '
    'values are replaced with the current fill value (or erased). Select '
    'mode is more powerful, but requires two steps to edit an image - '
    'you must first select some voxels, and then fill/erase them.',

    'OrthoEditProfile.mode' :
    'Switch between editing tools. The "Navigate" tool simply allows you to '
    'view the image and change the display location. The "Pencil" tool allows '
    'you to fill voxel values (in draw mode), or to manually select voxels '
    '(in select mode). The "Erase" tool allows you to erase voxel values (in '
    'draw mode), or to deselect voxels (in select mode). When select mode is '
    'enabled, the "select by intensity" tool allows you to select voxels '
    'based on their intensity. Click on a "seed" voxel, and all voxels with '
    'a similar intenstiy to that seed voxel will be selected.',

    'OrthoEditProfile.selectionSize' :
    'Size (in voxels) of the selection region when using the pencil or '
    'eraser tools.',

    'OrthoEditProfile.selectionIs3D' :
    'When using the pencil or eraser tools, choose between a 2D square '
    'selection in the plane of the active canvas, or a 3D cuboid. With '
    'the select by intensity tool, you can limit the selection search to '
    'the current 2D slice, or extend the search to the full 3D image.',

    'OrthoEditProfile.fillValue' :
    'Value to replace voxels with when drawing/filling.',

    'OrthoEditProfile.eraseValue' :
    'Value to replace voxels with when erasing.' ,

    'OrthoEditProfile.intensityThres' :
    'When using the select by intensity tool, this is the threshold used to '
    'determine whether or not a voxel should be selected. If the difference '
    'in intensity between the seed voxel and another voxel in the search '
    'space is less than or equal to this threshold, the voxel will be '
    'selected.',

    'OrthoEditProfile.intensityThresLimit' :
    'Upper limit for the intensity threshold. By default the upper intensity '
    'threshold limit is calculated from he image data range, but you can '
    'manually adjust it through this setting.',

    'OrthoEditProfile.localFill' :
    'When using the select by intensity tool, this setting will cause the '
    'search space to be limited to voxels which have a similar intensity '
    'to the seed voxel, and which are adjacent to another selected voxel.',

    'OrthoEditProfile.limitToRadius' :
    'When using the select by intensity tool, this setting will cause the '
    'search limited to a circle or sphere of the specified radius.',

    'OrthoEditProfile.searchRadius' :
    'When using the select by intensity tool, if the search is being limited '
    'to a radius, this setting allows you to specify the radius of the search '
    'circle/sphere.',

    'OrthoEditProfile.targetImage' : \
    'Choose the target image for edit operations. By default, when you '
    'fill/erase voxels, the currently selected image is modified. However, '
    'you can select a different image (of the same dimensions and resolution '
    'as the currently selected image) as the target for edit operations. This '
    'is most useful when selecting voxels by intensity - you can select voxels'
    'based on the values in the currently selected image, but then fill/erase '
    'that selection in another image.',
})


actions = TypeDict({
    'CanvasPanel.screenshot'        : 'Take a screenshot of the current scene',
    'CanvasPanel.OverlayDisplayPanel' : 'Show more overlay display settings',
    'CanvasPanel.CanvasSettingsPanel' : 'Show more view control settings',
    'CanvasPanel.OverlayInfoPanel'    : 'Show/hide the overlay '
                                        'information panel.',

    'OrthoPanel.toggleEditPanel' : 'Show/hide the edit settings panel',

    'PlotPanel.screenshot'       : 'Take a screenshot of the current plot.',
    'PlotPanel.importDataSeries' : 'Import data series from a text file.',
    'PlotPanel.exportDataSeries' : 'Export data series to a text file.',
    'PlotPanel.addDataSeries'    : 'Add (hold) data series '
                                   'from the current overlay.',
    'PlotPanel.removeDataSeries' : 'Remove the most recently '
                                   'added data series.',

    'TimeSeriesPanel.TimeSeriesControlPanel'  : 'Show/hide the time '
                                                'series control panel.',
    'TimeSeriesPanel.PlotListPanel'           : 'Show/hide the time '
                                                'series list panel.',
    'HistogramPanel.HistogramControlPanel'    : 'Show/hide the histogram '
                                                'control panel.',
    'HistogramPanel.PlotListPanel'            : 'Show/hide the histogram '
                                                'list panel.',

    'PowerSpectrumPanel.PowerSpectrumControlPanel'  : 'Show/hide the power '
                                                      'spectrum control '
                                                      'panel.',
    'PowerSpectrumPanel.PlotListPanel'              : 'Show/hide the power '
                                                      'spectrum list '
                                                      'panel.',

    'OrthoViewProfile.resetDisplay' : 'Reset the display on all canvases.',
    'OrthoViewProfile.centreCursor' : 'Reset location to centre of scene',

    'OrthoEditProfile.undo' :
    'Undo the most recent action. A history of changes to the selection, '
    'and to image data, are maintained. separate undo/redo histories are '
    'maintained for each image.',

    'OrthoEditProfile.redo' :
    'Redo the most recent undone action.',

    'OrthoEditProfile.createMask' :
    'Create an empty 3D mask image which has the same dimensions as the '
    'currently selected image.',

    'OrthoEditProfile.clearSelection' :
    'Clear the current selection - no voxels are selected.',

    'OrthoEditProfile.fillSelection' :
    'Fill selected voxels in the currently selected '
    'image with the current fill value. ',

    'OrthoEditProfile.eraseSelection' :
    'Set the value at all selected voxels in the currently selected '
    'image to zero.',

    'OrthoEditProfile.copyPasteData' :
    'Copy/paste data between images. Select some voxels in one image, and '
    'click this button to copy the voxel values to a clipboard. Then select '
    'another image (which has the same dimensions/resolution as the first '
    'image), and click this button again to paste the values. Shift+click the '
    'button to clear the clipboard.',

    'OrthoEditProfile.copyPasteSelection' :
    'Copy/paste 2D selections between slices. Draw a selection on one slice, '
    'and push this button to copy the selection from that slice. Then move .'
    'the cursor to a different slice, and push this button again to paste the '
    'selection into the new slice. Shift+click the button to clear the '
    'clipboard.',

    'Scene3DViewProfile.resetDisplay' :
    'Reset the zoom, pan, and rotation',

    # Items in the OverlayListPanel
    'ListItemWidget.save'  : 'Save this overlay to a file',
    'ListItemWidget.group' : 'Link some properties of this overlay '
                             'with other linked overlays (e.g. '
                             'volume)',

    'SampleLinePanel.screenshot' : 'Save the plot to a file',
    'SampleLinePanel.addDataSeries'    :
    'Add (hold) the most recently drawn data series.',
    'SampleLinePanel.removeDataSeries' :
    'Remove the most recently added data series.',

    'AnnotationPanel.loadAnnotations' :
    'Load annotations from a file',
    'AnnotationPanel.saveAnnotations' :
    'Save the currently displayed annotations to a file',
})


misc = TypeDict({
    'PlotControlPanel.labels' : 'X/Y axis labels.',
    'PlotControlPanel.logscale' :
    'If checked, a log (base 10) scale is used for the X/Y axis.',
    'PlotControlPanel.autoscale' :
    'If checked, the plot X/Y axis limits are automatically adjusted whenever '
    'the plot contents change.',
    'PlotControlPanel.invert' :
    'If checked, the plot is inverted along the X/Y axis.',
    'PlotControlPanel.scale' :
    'A constant scaling factor to apply to the data.',
    'PlotControlPanel.offset' :
    'A constant offset to apply to the data.',

    'PlotControlPanel.xlim'   : 'X axis data limits.',
    'PlotControlPanel.ylim'   : 'Y axis data limits.',

    'ResampleDialog.reference' :
    'Resample the image to the same dimensions as this reference.',
    'ResampleDialog.interpolation' :
    'The interpolation approach to use when resampling.',
    'ResampleDialog.origin' :
    'The origin common to the original space and the resampled space - .'
    'either the centre of the corner voxel, or the corner of the corner '
    'voxel.',
    'ResampleDialog.dtype' :
    'The data type of the resampled image. If you are resampling and '
    'interpolating an image with an integer data type, you may need to select '
    'a floating point type for the interpolation to have any effect.',
    'ResampleDialog.smoothing' :
    'If selected, the image data is smoothed with a gaussian filter before '
    'being resampled. This is to ensure that the values of the voxels in the '
    'image contribute more evenly to the values in the resampled image '
    'voxels. This setting has no effect when using nearest neighbour '
    'interpolation, and is only applied along axes which are being '
    'down-sampled.',
    'ResampleDialog.allVolumes' :
    'For images with more than three dimensions, this checkbox controls '
    'whether all volumes are resampled, or just the currently selected '
    'volume.',
})
